"""Time series analysis package."""

from enum import IntEnum
from typing import Any, Dict, List, Optional

import nbformat.v4 as nbfv4
import pandas as pd
from IPython.display import Markdown, display

from edvart.decorators import check_index_time_ascending
from edvart.report_sections.section_base import ReportSection, Verbosity
from edvart.report_sections.timeseries_analysis import (
    Autocorrelation,
    BoxplotsOverTime,
    FourierTransform,
    RollingStatistics,
    SeasonalDecomposition,
    ShortTimeFT,
    StationarityTests,
    TimeSeriesLinePlot,
)


class TimeseriesAnalysis(ReportSection):
    """Generates the Timeseries analysis section of the report.

    Contains an enum TimeseriesAnalysisSubsection of possible subsections.

    Parameters
    ----------
    subsections : List[TimeseriesAnalysisSubsection], optional
        List of subsections to include.
        All subsection in TimeseriesAnalysisSubsection are included by default, except for
        FourierTransform, which is only included if `sampling_rate` is set and ShortTimeFT, which
        is only included if `sampling_rate` and `stft_window_size` are both set.
    verbosity : Verbosity
        Generated code verbosity global to the Overview sections.
        If subsection verbosities are None, then they will be overridden by this parameter.
    columns : List[str], optional
        Columns to include in timeseries analysis. Each column is treated as a separate time series.
        All columns are used by default.
    verbosity_series_line_plot : Verbosity, optional
        Time series line plot subsection code verbosity.
    verbosity_rolling_statistics: Verbosity, optional
        Rolling statistics interactive plot subsection code verbosity.
    verbosity_boxplots_over_time: Verbosity, optional
        Boxplots grouped over time intervals subsection code verbosity.
    verbosity_seasonal_decomposition: Verbosity, optional
        Seasonal decomposition subsection code verbosity.
    verbosity_stationarity_tests: Verbosity, optional
        Stationarity tests subsection code verbosity.
    verbosity_autocorrelation: Verbosity, optional
        Autocorrelation and partial autocorrelation plot subsection code verbosity.
    verbosity_fourier_transform: Verbosity, optional
        Discrete Fourier transform plot subsection code verbosity.
    verbosity_short_time_ft: Verbosity, optional
        Short-time discrete Fourier transform plot subsection code verbosity.
    sampling_rate: int, optional
        Sampling rate of the time-series, i.e., how many samples form one period. For example,
        if your time-series contains hourly data and want to investigate daily frequencies, use 24.
        If not set, Fourier transform and Short-time Fourier transform will not be included.
    stft_window_size: int
        Window size for Short-time Fourier transform, which will not be included if this parameter
        is not set.
    """

    # pylint: disable=invalid-name
    class TimeseriesAnalysisSubsection(IntEnum):
        """Enum of all implemented timeseries analysis subsections."""

        TimeSeriesLinePlot = 0
        RollingStatistics = 1
        BoxplotsOverTime = 2
        SeasonalDecomposition = 3
        StationarityTests = 4
        Autocorrelation = 5
        FourierTransform = 6
        ShortTimeFT = 7

        def __str__(self):
            return self.name

    def __init__(
        self,
        subsections: Optional[List[TimeseriesAnalysisSubsection]] = None,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        verbosity_time_series_line_plot: Optional[Verbosity] = None,
        verbosity_rolling_statistics: Optional[Verbosity] = None,
        verbosity_boxplots_over_time: Optional[Verbosity] = None,
        verbosity_seasonal_decomposition: Optional[Verbosity] = None,
        verbosity_stationarity_tests: Optional[Verbosity] = None,
        verbosity_autocorrelation: Optional[Verbosity] = None,
        verbosity_fourier_transform: Optional[Verbosity] = None,
        verbosity_short_time_ft: Optional[Verbosity] = None,
        sampling_rate: Optional[int] = None,
        stft_window_size: Optional[int] = None,
    ):
        self.sampling_rate = sampling_rate
        self.stft_window_size = stft_window_size

        verbosity_time_series_line_plot = verbosity_time_series_line_plot or verbosity
        verbosity_rolling_statistics = verbosity_rolling_statistics or verbosity
        verbosity_boxplots_over_time = verbosity_boxplots_over_time or verbosity
        verbosity_seasonal_decomposition = (
            verbosity_seasonal_decomposition
            if verbosity_seasonal_decomposition is not None
            else verbosity
        )
        verbosity_stationarity_tests = verbosity_stationarity_tests or verbosity
        verbosity_autocorrelation = verbosity_autocorrelation or verbosity
        verbosity_fourier_transform = verbosity_fourier_transform or verbosity
        verbosity_short_time_ft = verbosity_short_time_ft or verbosity

        subsec = TimeseriesAnalysis.TimeseriesAnalysisSubsection

        verbosities = {
            subsec.TimeSeriesLinePlot: verbosity_time_series_line_plot,
            subsec.RollingStatistics: verbosity_rolling_statistics,
            subsec.BoxplotsOverTime: verbosity_boxplots_over_time,
            subsec.SeasonalDecomposition: verbosity_seasonal_decomposition,
            subsec.StationarityTests: verbosity_stationarity_tests,
            subsec.Autocorrelation: verbosity_autocorrelation,
            subsec.FourierTransform: verbosity_fourier_transform,
            subsec.ShortTimeFT: verbosity_short_time_ft,
        }

        enum_to_implementation = {
            subsec.TimeSeriesLinePlot: TimeSeriesLinePlot(verbosity_time_series_line_plot, columns),
            subsec.RollingStatistics: RollingStatistics(verbosity_rolling_statistics, columns),
            subsec.BoxplotsOverTime: BoxplotsOverTime(verbosity_boxplots_over_time, columns),
            subsec.SeasonalDecomposition: SeasonalDecomposition(
                verbosity_seasonal_decomposition, columns
            ),
            subsec.StationarityTests: StationarityTests(verbosity_stationarity_tests, columns),
            subsec.Autocorrelation: Autocorrelation(verbosity_autocorrelation, columns),
        }

        if sampling_rate is not None:
            enum_to_implementation[subsec.FourierTransform] = FourierTransform(
                sampling_rate, verbosity_fourier_transform, columns
            )
            if stft_window_size is not None:
                enum_to_implementation[subsec.ShortTimeFT] = ShortTimeFT(
                    sampling_rate, stft_window_size, verbosity_short_time_ft, columns
                )

        # By default use all subsections, FT and STFT only if required parameters specified
        if subsections is None:
            subsections_all = list(enum_to_implementation.keys())
        else:
            subsections_all = subsections

        # Store subsections with LOW verbosity
        self.subsections_low_verbosity = [
            sub for sub in subsections_all if verbosities[sub] == Verbosity.LOW
        ]

        if len(self.subsections_low_verbosity) == len(subsections_all) and subsections is None:
            self.subsections_low_verbosity = None

        if subsections is None:
            subsections_implementations = list(enum_to_implementation.values())
        else:
            if sampling_rate is None:
                if subsec.FourierTransform in subsections:
                    raise ValueError("Need to set a `sampling_rate` to plot Fourier transform.")
                if subsec.ShortTimeFT in subsections:
                    raise ValueError(
                        "Need to set a `sampling_rate` to plot Short-time Fourier transform."
                    )
            if stft_window_size is None and subsec.ShortTimeFT in subsections:
                raise ValueError(
                    "Need to set an `stft_window_size` to plot Short-time Fourier transform."
                )
            subsections_implementations = [enum_to_implementation[sub] for sub in subsections]

        super().__init__(subsections_implementations, verbosity, columns)

    @property
    def name(self) -> str:
        return "Timeseries analysis"

    @staticmethod
    @check_index_time_ascending
    def timeseries_analysis(
        df: pd.DataFrame,
        subsections: Optional[List[TimeseriesAnalysisSubsection]] = None,
        columns: Optional[List[str]] = None,
        sampling_rate: Optional[int] = None,
        stft_window_size: Optional[int] = None,
    ) -> None:
        """Generate timeseries analysis for df.

        Parameters
        ----------
        df : pd.DataFrame
            Data to be analyzed.
        subsections : List[TimeseriesAnalysisSubsection], optional
            Subsections to include in the analysis. All subsections are included by default.
        columns : List[str], optional
            Subset of columns of df to consider in timeseries analysis.
            All columns are used by default.
        sampling_rate : int, optional
            Sampling rate of the time-series, i.e., how many samples form one period. For example,
            if your timeseries contains hourly data and you want to investigate daily frequencies,
            use 24.
            If not set, Fourier transform and Short-time Fourier transform will not be included.
        stft_window_size : int, optional
            Window size for Short-time Fourier transform. Short-time Fourier transform will not be
            included if this parameter is not set.

        Raises
        ------
        ValueError
            If the input data is not indexed by time in ascending order.
        """
        if columns is not None:
            df = df[columns]

        timeseries_analysis = TimeseriesAnalysis(
            subsections=subsections,
            verbosity=Verbosity.LOW,
            columns=columns,
            sampling_rate=sampling_rate,
            stft_window_size=stft_window_size,
        )

        for sub in timeseries_analysis.subsections:
            sub.show(df)

    def add_cells(self, cells: List[Dict[str, Any]]) -> None:
        """Add cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)

        if self.verbosity == Verbosity.LOW:
            subsec = TimeseriesAnalysis.TimeseriesAnalysisSubsection
            code = "timeseries_analysis(df=df"

            if self.subsections_low_verbosity is not None:
                arg_subsections_names = [
                    f"TimeseriesAnalysis.TimeseriesAnalysisSubsection.{str(sub)}"
                    for sub in self.subsections_low_verbosity
                ]
                code += f", subsections={arg_subsections_names}".replace("'", "")

            stft_included_or_empty = (
                self.subsections_low_verbosity is None
                or subsec.ShortTimeFT in self.subsections_low_verbosity
            )
            include_sampling_rate = self.sampling_rate is not None and (
                stft_included_or_empty or subsec.FourierTransform in self.subsections_low_verbosity
            )
            if include_sampling_rate:
                code += f", sampling_rate={self.sampling_rate}"
                if self.stft_window_size is not None and stft_included_or_empty:
                    code += f", stft_window_size={self.stft_window_size}"

            if self.columns is not None:
                code += f", columns={self.columns}"
            code += ")"
            cells.append(nbfv4.new_code_cell(code))

            for sub in self.subsections:
                if sub.verbosity > Verbosity.LOW:
                    sub.add_cells(cells)
        else:
            super().add_cells(cells)

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"]
        """
        if self.verbosity == Verbosity.LOW:
            imports = {
                "from edvart.report_sections.timeseries_analysis import TimeseriesAnalysis\n"
                "timeseries_analysis = TimeseriesAnalysis.timeseries_analysis"
            }
            for sub in self.subsections:
                if sub.verbosity > Verbosity.LOW:
                    imports.update(sub.required_imports())

            return list(imports)
        return super().required_imports()

    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output of this section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output.
        """
        display(Markdown(self.get_title(section_level=1)))
        super().show(df)



"""Bivariate analysis section package."""

from enum import IntEnum
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import numpy as np
import pandas as pd
import seaborn as sns
from IPython.display import Markdown, display

from edvart import utils
from edvart.data_types import is_boolean, is_categorical, is_numeric
from edvart.report_sections.code_string_formatting import get_code, total_dedent
from edvart.report_sections.section_base import ReportSection, Section, Verbosity


class BivariateAnalysis(ReportSection):
    """Generates the Bivariate analysis section of the report.

    Contains an enum BivariateAnalysisSubsection of possible subsections.

    Parameters
    ----------
    subsections : List[BivariateAnalysisSubsection], optional
        List of subsections to include.
        All subsection in BivariateAnalysisSubsection are included by default.
    verbosity : Verbosity (default = Verbosity.LOW)
        Generated code verbosity global to the Bivariate analysis sections.
        If subsection verbosities are None, then they will be overridden by this parameter.
    columns : List[str], optional
        Columns on which to do bivariate analysis.
        If none of columns_x, columns_y and columns_pairs is specified bivariate analysis is
        performed on all pairs of `columns`.
        Ignored if `columns_x` and `columns_y` is specified. Ignored in contingency table
        if `columns_x` and `columns_y` or `columns_pairs` is specified.
        All columns are used by default.
    columns_x : List[str], optional
        If specified, correlations and pairplots are performed on the cartesian product of
        `columns_x` and `columns_y`.
        If `columns_x` is specified, then `columns_y` must also be specified.
    columns_y : List[str], optional
        If specified, correlations and pairplots are performed on the cartesian product of
        `columns_x` and `columns_y`.
        If `columns_y` is specified, then `columns_x` must also be specified.
    columns_pairs : List[str], optional
        List of columns pairs on which to perform bivariate analysis.
        Used primarily in contingency tables.
        If specified, `columns`, `columns_x` and `columns_y` are ignored in contingency tables.
        Ignored in pairplots and correlations unless `columns_pairs` is specified and none of
        `columns`, `columns_x`, `columns_y` is specified. In that case, the first elements of each
        pair are treated as `columns_x` and the second elements as `columns_y` in pairplots and
        correlations.
    verbosity_correlations : Verbosity, optional
        Correlation plots subsection code verbosity.
    verbosity_pairplot: Verbosity, optional
        Pairplot subsection code verbosity.
    verbosity_contingency_table: Verbosity, optional
        Contingency table subsection code verbosity.
    color_col : str, optional
        Name of column according to use for coloring of the bivariate analysis subsections.
        Coloring is currently supported in pairplot.

    Raises
    ------
    ValueError
        If exactly one of `columns_x`, `columns_y` is specified.
    """

    # pylint:disable=invalid-name
    class BivariateAnalysisSubsection(IntEnum):
        """Enum of all implemented bivariate analysis subsections."""

        CorrelationPlot = 0
        PairPlot = 1
        ContingencyTable = 2

        def __str__(self):
            return self.name

    def __init__(
        self,
        subsections: Optional[List[BivariateAnalysisSubsection]] = None,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        columns_pairs: Optional[List[Tuple[str, str]]] = None,
        verbosity_correlations: Optional[Verbosity] = None,
        verbosity_pairplot: Optional[Verbosity] = None,
        verbosity_contingency_table: Optional[Verbosity] = None,
        color_col: Optional[str] = None,
    ):
        verbosity_correlations = verbosity_correlations or verbosity
        verbosity_pairplot = verbosity_pairplot or verbosity
        verbosity_contingency_table = verbosity_contingency_table or verbosity

        subsec = BivariateAnalysis.BivariateAnalysisSubsection

        # Store subsection verbosities
        verbosities = {
            subsec.CorrelationPlot: verbosity_correlations,
            subsec.PairPlot: verbosity_pairplot,
            subsec.ContingencyTable: verbosity_contingency_table,
        }

        # By default use all subsections
        if subsections is None:
            subsections_all = list(BivariateAnalysis.BivariateAnalysisSubsection)
        else:
            subsections_all = subsections

        # Store subsections with LOW verbosity
        self.subsections_low_verbosity = [
            sub for sub in subsections_all if verbosities[sub] == Verbosity.LOW
        ]

        if len(self.subsections_low_verbosity) == len(subsections_all) and subsections is None:
            self.subsections_low_verbosity = None

        if (columns_x is None) != (columns_y is None):
            raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
        # For analyses which do not take columns_pairs, prepare columns_x and columns_y in case
        # columns_pairs is the only parameter specified
        if columns is None and columns_x is None and columns_pairs is not None:
            columns_x_no_pairs = [pair[0] for pair in columns_pairs]
            columns_y_no_pairs = [pair[1] for pair in columns_pairs]
        else:
            columns_x_no_pairs = columns_x
            columns_y_no_pairs = columns_y
        enum_to_implementation = {
            subsec.CorrelationPlot: CorrelationPlot(
                verbosity_correlations, columns, columns_x_no_pairs, columns_y_no_pairs
            ),
            subsec.PairPlot: PairPlot(
                verbosity_pairplot,
                columns,
                columns_x_no_pairs,
                columns_y_no_pairs,
                color_col=color_col,
            ),
            subsec.ContingencyTable: ContingencyTable(
                verbosity_contingency_table, columns, columns_x, columns_y, columns_pairs
            ),
        }

        subsections_implementations = [enum_to_implementation[sub] for sub in subsections_all]
        super().__init__(subsections_implementations, verbosity, columns)

        self.columns_x = columns_x
        self.columns_y = columns_y
        self.columns_pairs = columns_pairs
        self.color_col = color_col

    @property
    def name(self) -> str:
        return "Bivariate analysis"

    @staticmethod
    def bivariate_analysis(
        df: pd.DataFrame,
        subsections: Optional[List[BivariateAnalysisSubsection]] = None,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        columns_pairs: Optional[List[Tuple[str, str]]] = None,
        color_col: Optional[str] = None,
    ) -> None:
        """Generates bivariate analysis for df.

        Parameters
        ----------
        df : pd.DataFrame
            Data to be analyzed
        subsections : List[BivariateAnalysisSubsection], optional
            Subsections to include in the analysis. All subsections are included by default.
        columns : List[str], optional
            Columns on which to do bivariate analysis.
            If none of columns_x, columns_y and columns_pairs is specified bivariate analysis is
            performed on all pairs of `columns`.
            Ignored if `columns_x` and `columns_y` is specified. Ignored in contingency table
            if `columns_x` and `columns_y` or `columns_pairs` is specified.
            All columns are used by default.
        columns_x : List[str], optional
            If specified, correlations and pairplots are performed on the cartesian product of
            `columns_x` and `columns_y`.
            If `columns_x` is specified, then `columns_y` must also be specified.
        columns_y : List[str], optional
            If specified, correlations and pairplots are performed on the cartesian product of
            `columns_x` and `columns_y`.
            If `columns_y` is specified, then `columns_x` must also be specified.
        columns_pairs : List[str], optional
            List of columns pairs on which to perform bivariate analysis.
            Used primarily in contingency tables.
            If specified, `columns`, `columns_x` and `columns_y` are ignored in contingency tables.
            Ignored in pairplots and correlations unless `columns_pairs` is specified and none of
            `columns`, `columns_x`, `columns_y` is specified. In that case, the first elements of
            each pair are treated as `columns_x` and the second elements as `columns_y` in pairplots
            and correlations.
        color_col : str, optional
            Name of column to use for coloring of the bivariate analysis subsections.
            Coloring is currently supported in pairplot.
        """
        bivariate_analysis = BivariateAnalysis(
            subsections=subsections,
            verbosity=Verbosity.LOW,
            columns=columns,
            columns_x=columns_x,
            columns_y=columns_y,
            columns_pairs=columns_pairs,
            color_col=color_col,
        )

        for sub in bivariate_analysis.subsections:
            sub.show(df)

    def add_cells(self, cells: List[Dict[str, Any]]) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)
        if self.verbosity == Verbosity.LOW:
            code = "bivariate_analysis(df=df"
            if self.subsections_low_verbosity is not None:
                arg_subsections_names = [
                    f"BivariateAnalysis.BivariateAnalysisSubsection.{str(sub)}"
                    for sub in self.subsections_low_verbosity
                ]

                code += f", subsections={arg_subsections_names}".replace("'", "")
            if self.columns_x is not None:
                code += f", columns_x={self.columns_x}"
                code += f", columns_y={self.columns_y}"
            elif self.columns is not None:
                code += f", columns={self.columns}"
            if self.columns_pairs is not None:
                code += f", columns_pairs={self.columns_pairs}"
            if self.color_col is not None:
                code += f", color_col='{self.color_col}'"
            code += ")"
            cells.append(nbfv4.new_code_cell(code))
            for sub in self.subsections:
                if sub.verbosity > Verbosity.LOW:
                    sub.add_cells(cells)
        else:
            super().add_cells(cells)

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np']
        """
        if self.verbosity != Verbosity.LOW:
            return super().required_imports()

        imports = {
            "from edvart.report_sections.bivariate_analysis import BivariateAnalysis\n"
            "bivariate_analysis = BivariateAnalysis.bivariate_analysis"
        }
        for subsec in self.subsections:
            if subsec.verbosity > Verbosity.LOW:
                imports.update(subsec.required_imports())

        return list(imports)

    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output of this section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output.
        """
        display(Markdown(self.get_title(section_level=1)))
        super().show(df)


class CorrelationPlot(Section):
    """Generates the Correlation plot subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns on which to plot pair-wise correlation plot.
        If `columns_x` and `columns_y` are unspecified, then analysis is performed on all pairs
        of `columns`. Otherwise ignored.
        All columns are used by default.
    columns_x : List[str], optional
        If specified, correlation is plotted on the cartesian product
        of `columns_x` and `columns_y`.
        If `columns_x` is specified, then `columns_y` must also be specified.
    columns_y : List[str], optional
        If specified, correlation is plotted on the cartesian product
        of `columns_x` and `columns_y`.
        If `columns_y` is specified, then `columns_x` must also be specified.

    Raises
    ------
    ValueError
        If exactly one of `columns_x`, `columns_y` is specified.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
    ):
        if (columns_x is None) != (columns_y is None):
            raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
        self.columns_x = columns_x
        self.columns_y = columns_y
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Correlation Plot"

    @staticmethod
    def _get_columns_x_y(
        df: pd.DataFrame,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
    ) -> Tuple[List[str], List[str]]:
        """
        Helper function to get columns_x and columns_y from provided parameters and dataset.

        Parameters
        ----------
        df : pd.DataFrame
            Dataset to analyze.
        columns : List[str], optional
            Specified columns parameter.
        columns_x : List[str], optional
            Specified columns_x parameter.
        columns_y : List[str], optional
            Specified columns_x parameter.


        Returns
        -------
        columns_x, columns_y: Tuple[List[str], List[str]]
            Resolved columns_x and columns_y.
        """
        if (columns_x is None) != (columns_y is None):
            raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
        if columns_x is None:
            if columns is None:
                columns = list(df.columns)
            columns_x = columns_y = columns
        columns_x = [col for col in columns_x if is_numeric(df[col])]
        columns_y = [col for col in columns_y if is_numeric(df[col])]

        return columns_x, columns_y

    @staticmethod
    def default_correlations() -> Dict[str, Callable[[pd.DataFrame], pd.DataFrame]]:
        """
        Default dictionary of bivariate statistics that will be calculated for numerical columns.

        Returns
        -------
        dict
            A dictionary assigning correlation functions to correlation names.
            Dictionary signature: 'CorrelationName': corr_func
        """
        return {
            "pearson": utils.pearson,
            "spearman": utils.spearman,
            "kendall": utils.kendall,
        }

    @staticmethod
    def plot_correlation(
        df: pd.DataFrame,
        corr_name: str,
        corr_func: Callable[[pd.DataFrame], pd.DataFrame],
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        size_factor: float = 0.7,
        font_size: float = 15,
        color_map: Optional[Any] = None,
    ) -> None:
        """Plots a correlation heatmap.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to plot correlation.
        corr_name : str
            Name of correlation being plotted.
        corr_func : Callable[[np.ndarray, Optional[List[Tuple[str, str]]]], pd.DataFrame]
            Correlation function to be used.
        columns : Optional[List[str]]
            List of columns of df to analyze. All numeric columns of df are used by default.
        size_factor: float (default = 0.7)
            Size of each cell in the table.
        font_size : float (default = 15)
            Size of axis labels of the correlation plot.
        color_map : Any, optional
            Color map compatible with matplotlib/seaborn to use for the correlation plot.
            A divergent red-blue color map is used by default.

        Raises
        ------
        ValueError
            If exactly one of `columns_x`, `columns_y` is specified.
        """
        columns_x, columns_y = CorrelationPlot._get_columns_x_y(df, columns, columns_x, columns_y)
        if color_map is None:
            color_map = sns.diverging_palette(10, 220, sep=80, n=100)

        # show header
        display(Markdown(f"### {corr_name.capitalize()} Correlation"))

        # calculate correlation between columns
        corr = corr_func(df)
        corr = corr.loc[columns_x, columns_y]

        # plot correlation heatmap
        ax = sns.heatmap(
            corr, cmap=color_map, vmin=-1, vmax=1, square=True, xticklabels=True, yticklabels=True
        )

        # set axes
        ax.set_xticklabels(ax.get_xticklabels(), fontsize=font_size, rotation=90)
        ax.set_yticklabels(ax.get_yticklabels(), fontsize=font_size, rotation=0)
        ax.figure.set_size_inches(
            len(columns_x) * size_factor + 1, len(columns_y) * size_factor + 1
        )

        plt.show()

    @staticmethod
    def plot_correlations(
        df: pd.DataFrame,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        pearson: bool = True,
        spearman: bool = True,
        kendall: bool = True,
    ) -> None:
        """Plots multiple correlations.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to plot correlations.
        columns : Optional[List[str]]
            List of columns of df to analyze. All numeric columns of df are used by default.
        pearson : bool (default = True)
            If True, Pearson correlation will be plotted.
        spearman : bool (default = True)
            If True, Spearman correlation will be plotted.
        kendall : bool (default = True)
            If True, Kendall correlation will be plotted.
        """
        correlations = CorrelationPlot.default_correlations()

        # remove correlations which are not to be plotted
        if not pearson:
            correlations.pop("pearson", None)
        if not spearman:
            correlations.pop("spearman", None)
        if not kendall:
            correlations.pop("kendall", None)

        for corr_name, corr_func in correlations.items():
            CorrelationPlot.plot_correlation(
                df, corr_name, corr_func, columns, columns_x, columns_y
            )

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                total_dedent(
                    """
                    from edvart.report_sections.bivariate_analysis import CorrelationPlot
                    plot_correlations = CorrelationPlot.plot_correlations
                    """
                )
            ]
        return [
            "import numpy as np",
            "import seaborn as sns",
            "from IPython.display import display, Markdown",
            "import matplotlib.pyplot as plt",
            "%matplotlib inline",
            "from edvart import utils",
            "from edvart.data_types import is_numeric",
        ]

    def add_cells(self, cells: List[Dict[str, Any]]) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "plot_correlations(df=df"
        if self.columns_x is not None:
            default_call += f", columns_x={self.columns_x}"
            default_call += f", columns_y={self.columns_y}"
        elif self.columns is not None:
            default_call += f", columns={self.columns}"

        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(CorrelationPlot.default_correlations)
                + "\n\n"
                + get_code(CorrelationPlot._get_columns_x_y)
                + "\n\n"
                + get_code(CorrelationPlot.plot_correlation).replace("CorrelationPlot.", "")
                + "\n\n"
                + get_code(CorrelationPlot.plot_correlations).replace("CorrelationPlot.", "")
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates correlation plots in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        CorrelationPlot.plot_correlations(
            df=df, columns=self.columns, columns_x=self.columns_x, columns_y=self.columns_y
        )


class PairPlot(Section):
    """Generates the Pairplot subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns on which to plot the pairplot.
        If `columns_x` and `columns_y` are unspecified, then analysis is performed on all pairs
        of `columns`. Otherwise ignored.
        All columns are used by default.
    columns_x : List[str], optional
        If specified, correlation is plotted
        on the cartesian product of `columns_x` and `columns_y`.
        If `columns_x` is specified, then `columns_y` must also be specified.
    columns_y : List[str], optional
        If specified, correlation is plotted
        on the cartesian product of `columns_x` and `columns_y`.
        If `columns_y` is specified, then `columns_x` must also be specified.
    color_col : str, optional
        Name of column according to use for coloring of points and histogram in the pairplot.

    Raises
    ------
    ValueError
        If exactly one of `columns_x`, `columns_y` is specified.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        color_col: Optional[str] = None,
    ):
        if (columns_x is None) != (columns_y is None):
            raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
        self.columns_x = columns_x
        self.columns_y = columns_y
        self.color_col = color_col
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Pairplot"

    @staticmethod
    def plot_pairplot(
        df: pd.DataFrame,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        allow_categorical: bool = False,
        color_col: Optional[str] = None,
    ) -> None:
        """Plot a pairplot for each pair of columns.

        Parameters
        ----------
        df : pd.DataFrame
            Data frame for which to plot pairplot.
        columns : Union[List[Tuple[str, str]] or List[str]], optional
            Which columns to plot pairplot for.
            All columns that are not categorical and are not boolean are used by default.
        columns_x : List[str], optional
            If specified, correlation is plotted
            on the cartesian product of `columns_x` and `columns_y`.
            If `columns_x` is specified, then `columns_y` must also be specified.
        columns_y : List[str], optional
            If specified, correlation is plotted
            on the cartesian product of `columns_x` and `columns_y`.
            If `columns_y` is specified, then `columns_x` must also be specified.
        allow_categorical : bool (default = False)
            Whether to allow plotting of categorical columns. If False (default), then even
            explicitly specified columns will be excluded. If True, categorical columns are still
            excluded by default, unless explicitly specified via columns/columns_x/columns_y.
        color_col : str, optional
            Name of column according to use for coloring of points and histogram in the pairplot.

        Raises
        ------
        ValueError
            If exactly one of `columns_x`, `columns_y` is specified.
        """

        def include_column(col: str) -> bool:
            return not is_categorical(df[col]) and not is_boolean(df[col])

        if (columns_x is None) != (columns_y is None):
            raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
        if columns_x is None:
            if columns is None:
                columns = list(filter(include_column, df.columns))
            columns_x = columns
            columns_y = columns
        if not allow_categorical:
            columns_x = list(filter(include_column, columns_x))
            columns_y = list(filter(include_column, columns_y))
        sns.pairplot(df, x_vars=columns_x, y_vars=columns_y, hue=color_col)
        plt.tight_layout()
        plt.show()

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                total_dedent(
                    """
                    from edvart.report_sections.bivariate_analysis import PairPlot
                    plot_pairplot = PairPlot.plot_pairplot
                    """
                )
            ]
        return [
            "from edvart.data_types import is_categorical, is_boolean",
            "import matplotlib.pyplot as plt",
            "import seaborn as sns",
        ]

    def add_cells(self, cells: List[Dict[str, Any]]) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "plot_pairplot(df=df"
        if self.columns_x is not None:
            default_call += f", columns_x={self.columns_x}"
            default_call += f", columns_y={self.columns_y}"
        elif self.columns is not None:
            default_call += f", columns={self.columns})"
        if self.color_col is not None:
            default_call += f", color_col='{self.color_col}'"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = get_code(PairPlot.plot_pairplot) + "\n\n" + default_call

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates pairplot in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        PairPlot.plot_pairplot(
            df=df,
            columns=self.columns,
            columns_x=self.columns_x,
            columns_y=self.columns_y,
            color_col=self.color_col,
        )


class ContingencyTable(Section):
    """Generates the pairwise contingency tables subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns on which to show contingency tables.
        If `columns_x` and `columns_y` and `columns_pairs` are unspecified,
        then analysis performed on all pairs of `columns`. Otherwise ignored.
        Columns which contain only null values are always excluded. All columns are used by default.
    columns_x : List[str], optional
        If specified, contingency tables are plotted for each pair
        in the cartesian product of `columns_x` and `columns_y`.
        If `columns_x` is specified, then `columns_y` must also be specified.
        Columns which contain only null values are always excluded.
        Ignored if `columns_pairs` is specified.
    columns_y : List[str], optional
        If specified, contingency tables are plotted for each pair
        in the cartesian product of `columns_x` and `columns_y`.
        If `columns_y` is specified, then `columns_x` must also be specified.
        Columns which contain only null values are always excluded.
        Ignored if `columns_pairs` is specified.
    columns_pairs : List[Tuple[str, str]], optional
        If specified, contingency tables are plotted for exactly the specified pairs.
        Columns which contain only null values are always excluded, i.e. if at least one of the
        columns in a pair is excluded, then is excluded.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        columns_pairs: Optional[List[Tuple[str, str]]] = None,
    ):
        super().__init__(verbosity, columns)
        self.columns_x = columns_x
        self.columns_y = columns_y
        self.columns_pairs = columns_pairs

    @property
    def name(self) -> str:
        return "Contingency table"

    @staticmethod
    def contingency_tables(
        df: pd.DataFrame,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        columns_pairs: Optional[List[Tuple[str, str]]] = None,
        table_threshold: int = 30,
    ) -> None:
        """Display a contingency table for each pairs of columns.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to create a contingency table.
        columns : List[str], optional
            Which columns to generate pair-wise contingency tables for.
            All columns which contain more than 1 unique value are used by default.
            Columns which contain only null values are always excluded.
        columns_x : List[str], optional
            If specified, contingency tables are plotted for each pair
            in the cartesian product of `columns_x` and `columns_y`.
            If `columns_x` is specified, then `columns_y` must also be specified.
            Ignored if `columns_pairs` is specified.
        columns_y : List[str], optional
            If specified, contingency tables are plotted for each pair
            in the cartesian product of `columns_x` and `columns_y`.
            If `columns_y` is specified, then `columns_x` must also be specified.
            Ignored if `columns_pairs` is specified.
        columns_pairs : List[Tuple[str, str]], optional
            If specified, contingency tables are plotted for exactly the specified pairs.
        table_threshold : int (default = 30)
            Maximum number of unique values for a column to be used in contingency table.
            If non-positive, no columns are excluded according to this criterion.
        """

        def include_column(col: str) -> bool:
            return (table_threshold < 0 or df[col].nunique() <= table_threshold) and (
                not df[col].isnull().all()
            )

        def include_column_pair(col_pair: Tuple[str, str]) -> bool:
            if col_pair[0] == col_pair[1]:
                return False
            return include_column(col_pair[0]) and include_column(col_pair[1])

        if (columns_x is None) != (columns_y is None):
            raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
        if columns_pairs is None:
            if columns_x is None:
                if columns is None:
                    columns = [col for col in df.columns if df[col].nunique() > 1]
                columns_x = columns
                columns_y = columns
            columns_pairs = [(col_x, col_y) for col_x in columns_x for col_y in columns_y]
        columns_pairs = list(filter(include_column_pair, columns_pairs))

        for column1, column2 in columns_pairs:
            ContingencyTable.contingency_table(df, column1, column2)

    @staticmethod
    def contingency_table(
        df: pd.DataFrame,
        columns1: Union[str, List[str]],
        columns2: Union[str, List[str]],
        include_total: bool = True,
        hide_zeros: bool = True,
        scaling_func: Callable[[np.ndarray], np.ndarray] = np.cbrt,
        colormap: Any = "Blues",
        size_factor: float = 0.7,
        fontsize: float = 15,
    ) -> None:
        """
        Show a colored contingency table for the two specified columns or lists of columns.

        Parameters
        ----------
        df : pd.DataFrame
            Data to analyze.
        columns1 : Union[str, List[str]]
            Name of column or list of column names to use in the vertical axis of the table.
        columns2 : Union[str, List[str]]
            Name of column or list of column names to use in the horizontal axis of the table.
        include_total : bool (default = True)
            Whether to include marginal counts.
        hide_zeros : bool (default = True)
            Whether to hide zero values in the table for better readability.
        scaling_func : Callable[["array-like"], "array-like"]
            Function to scale the values for the purpose of coloring for smaller spread.
            Cube root is used by default.
        colormap : Any (default = "Blues")
            Colormap compatible with matplotlib/seaborn.
        size_factor : float (default = 0.7)
            Size of each cell in the table.
        fontsize : float (default = 15)
            Size of the font for axis labels.
        """
        if isinstance(columns1, str):
            columns1 = [columns1]
        if isinstance(columns2, str):
            columns2 = [columns2]
        table = pd.crosstab(
            [df[col] for col in columns1],
            [df[col] for col in columns2],
            margins_name="Total",
            margins=include_total,
        )

        annot = table.replace(0, "") if hide_zeros else table

        ax = sns.heatmap(
            scaling_func(table),
            annot=annot,
            fmt="",
            cbar=False,
            cmap=colormap,
            linewidths=0.1,
            xticklabels=1,
            yticklabels=1,
            annot_kws={"fontsize": fontsize},
            square=True,
        )
        ax.figure.set_size_inches(size_factor * len(table.columns), size_factor * len(table))

        ax.set_xticklabels(ax.get_xticklabels(), fontsize=fontsize)

        ax.set_ylabel(ax.get_ylabel(), fontsize=fontsize)
        ax.set_yticklabels(ax.get_yticklabels(), rotation=0, fontsize=fontsize)

        ax.xaxis.tick_top()
        ax.set_xlabel(ax.get_xlabel(), fontsize=fontsize)
        ax.xaxis.set_label_position("top")

        # Viusally separate the margins
        if include_total:
            ax.vlines(len(table.columns) - 1, ymin=0, ymax=len(table), color="grey")
            ax.hlines(len(table) - 1, xmin=0, xmax=len(table.columns), color="grey")

        plt.show()

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                total_dedent(
                    """
                    from edvart.report_sections.bivariate_analysis import ContingencyTable
                    contingency_tables = ContingencyTable.contingency_tables
                    """
                )
            ]
        return [
            "import itertools",
            "from edvart import utils",
            "import seaborn as sns",
            "import matplotlib.pyplot as plt",
        ]

    def add_cells(self, cells: List[Dict[str, Any]]) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "contingency_tables(df=df"
        if self.columns_pairs is not None:
            default_call += f", columns_pairs={self.columns_pairs}"
        elif self.columns_x is not None:
            default_call += f", columns_x={self.columns_x}"
            default_call += f", columns_y={self.columns_y}"
        elif self.columns is not None:
            default_call += f", columns={self.columns}"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(ContingencyTable.contingency_tables).replace("ContingencyTable.", "")
                + "\n\n"
                + get_code(ContingencyTable.contingency_table)
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates contingency table in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        ContingencyTable.contingency_tables(
            df=df,
            columns=self.columns,
            columns_x=self.columns_x,
            columns_y=self.columns_y,
            columns_pairs=self.columns_pairs,
        )



"""Univariate analysis package."""
from typing import Any, Callable, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import pandas as pd
import seaborn as sns
from IPython.display import HTML, Markdown, display

from edvart import utils
from edvart.data_types import DataType, infer_data_type
from edvart.pandas_formatting import add_html_heading, dict_to_html, format_number, subcells_html
from edvart.report_sections.code_string_formatting import code_dedent, get_code
from edvart.report_sections.section_base import Section, Verbosity


class UnivariateAnalysis(Section):
    """Generates univariate analysis section of the report.

    Parameters
    -----------
    df : pd.DataFrame
        Dataframe to be analyzed
    verbosity : Verbosity
        The verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns for which to do univariate analysis,
        all columns are used by default
    """

    def __init__(
        self,
        df: pd.DataFrame,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
    ):
        self.df = df
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Univariate Analysis"

    @staticmethod
    def default_descriptive_statistics():
        """
        Return default descriptive statistics.

        Returns
        -------
        dict
            Dictionary with keys as statistics names and values with functions.
        """
        return {
            "Number of unique values": utils.num_unique_values,
            "Sum": utils.sum_,
            "Mean": utils.mean,
            "Mode": utils.mode,
            "Standard deviation": utils.std,
            "Mean absolute deviation": utils.mad,
            "Median absolute deviation": utils.median_absolute_deviation,
            "Coefficient of variation": utils.coefficient_of_variation,
            "Kurtosis": utils.kurtosis,
            "Skewness": utils.skewness,
        }

    @staticmethod
    def default_quantile_statistics():
        """
        Return default quantile statistics.

        Returns
        -------
        dict
            Dictionary with keys as statistics names and values with functions.
        """
        return {
            "Minimum": utils.minimum,
            "Maximum": utils.maximum,
            "Q1": utils.quartile1,
            "Median": utils.median,
            "Q3": utils.quartile3,
            "Range": utils.value_range,
            "IQR": utils.iqr,
        }

    @staticmethod
    def histogram(
        series: pd.Series,
        bins: Optional[int] = None,
        density: bool = False,
        box_plot: bool = True,
        figsize: Tuple[float, float] = (20, 7),
        distplot_kwargs: Optional[Dict[str, Any]] = None,
        boxplot_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Visualizes distribution of series with a histogram.

        Parameters
        ----------
        series : pd.Series
            Numerical series.
        bins : int, optional
            Number of bins of the histogram. If None, the number of bins is inferred.
        density : bool (default = False)
            If True, the area of the histogram bars will sum up to 1.
        box_plot : bool (default = True)
            If True, a horizontal box plot will be added above the histogram to visualize quartiles.
        figsize : Tuple[float, float] (default = (20, 7))
            Size of the figure of the visualization.
        distplot_kwargs : Dict[str, Any], optional
            Additional keyword arguments passed to seaborn.distplot
        boxplot_kwargs : Dict[str, Any], optional
            Additional keyword arguments passed to seaborn.boxplot
        """
        if distplot_kwargs is None:
            distplot_kwargs = {}
        if boxplot_kwargs is None:
            boxplot_kwargs = {}

        series = series.dropna()

        if box_plot:
            _fig, (ax_box, ax_hist) = plt.subplots(
                nrows=2,
                sharex=True,
                gridspec_kw={"height_ratios": (0.15, 0.85)},
                figsize=figsize,
            )
            sns.boxplot(x=series, ax=ax_box, **boxplot_kwargs)
            sns.histplot(
                data=series,
                bins=bins or "auto",
                stat="density" if density else "count",
                ax=ax_hist,
                kde=False,
                **distplot_kwargs,
            )
            ax_box.set(xlabel="")
        else:
            plt.figure(figsize=figsize)
            sns.histplot(
                data=series,
                bins=bins or "auto",
                stat="density" if density else "count",
                kde=False,
                **distplot_kwargs,
            )
        plt.show()

    @staticmethod
    def bar_plot(
        series: pd.Series,
        relative_count: bool = False,
        figsize: Tuple[float, float] = (20, 7),
        plotting_threshold: int = 50,
        **bar_plot_args: Dict[str, Any],
    ) -> None:
        """Plots a bar plot visualizing frequencies of series elements.

        Parameters
        ----------
        series : pd.Series
            Categorical series.
        relative_count : bool
            If True, the frequencies will be normalized by the series length.
        figsize : Tuple[float, float]
            Size of the bar plot.
        plotting_threshold : int
            If the number of unique values in the series is greater than this, no plot is created
            instead a warning is issued.
        bar_plot_args : Dict[str, Any]
            Additional kwargs passed to pandas.Series.bar.
        """
        if series.nunique() > plotting_threshold:
            display(
                Markdown(
                    f"Number of unique values is greater than {plotting_threshold},"
                    " not plotting bar plot."
                )
            )
        else:
            value_counts = series.value_counts()
            value_counts = value_counts / len(series) if relative_count else value_counts
            value_counts.plot.bar(figsize=figsize, **bar_plot_args)
            plt.show()

    @staticmethod
    def numeric_statistics(
        series: pd.Series,
        descriptive_stats: Optional[Dict[str, Callable]] = None,
        quantile_stats: Optional[Dict[str, Callable]] = None,
        thousand_separator: str = " ",
    ) -> None:
        """Generates tables with statistics for numeric series.

        Parameters
        ----------
        series : pd.Series
            Numeric series.
        descriptive_stats : Dict[str, Callable], optional
            Descriptive statistics computed for the series.
            Dictionary format: {'statistic name': stat_func(pd.Series) -> float}
            If None, UnivariateAnalysis.default_descriptive_statistics is used.
        quantile_stats : Dict[str, Callable], optional
            Quantile statistics computed for the series.
            Dictionary format: {'statistic name': stat_func(pd.Series) -> float}
            If None, UnivariateAnalysis.default_quantile_statistics is used.
        thousand_separator : str
            Thousand separator for numbers in the tables, by default space.
        """
        if descriptive_stats is None:
            descriptive_stats = UnivariateAnalysis.default_descriptive_statistics()
        if quantile_stats is None:
            quantile_stats = UnivariateAnalysis.default_quantile_statistics()

        desc_stats = {
            stat_name: format_number(
                number=stat_func(series), thousand_separator=thousand_separator
            )
            for stat_name, stat_func in descriptive_stats.items()
        }
        quant_stats = {
            stat_name: format_number(
                number=stat_func(series), thousand_separator=thousand_separator
            )
            for stat_name, stat_func in quantile_stats.items()
        }
        # Render statistics tables side by side
        desc_stats_html = add_html_heading(dict_to_html(desc_stats), "Descriptive Statistics", 3)
        quant_stats_html = add_html_heading(dict_to_html(quant_stats), "Quantile Statistics", 3)
        html_table = [[desc_stats_html, quant_stats_html]]
        # Render
        display(HTML(subcells_html(html_table)))

    @staticmethod
    def top_most_frequent(series: pd.Series, n_top: int = 5) -> None:
        """Generates a table with top n most frequent values in series.

        Parameters
        ----------
        series : pd.Series
            Categorical series
        n_top : int
            The number of most frequent values to include in the table.
        """
        frequent_values = dict_to_html(utils.top_frequent_values(series, n_top=n_top))
        frequent_values_html = add_html_heading(frequent_values, "Most frequent values", 3)
        display(HTML(frequent_values_html))

    @staticmethod
    def univariate_analysis(df: pd.DataFrame, columns: Optional[List[str]] = None) -> None:
        """Generates univariate analysis for df.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe to be analyzed
        columns : List[str], optional
            Subset of df columns to analyze, by default all columns of df are used
        """
        if columns is not None:
            df = df[columns]

        for col in df.columns:
            data_type = infer_data_type(df[col])
            data_type_name = str(data_type)
            display(Markdown(f"## *{col} - {data_type_name}*"))
            if data_type in (DataType.CATEGORICAL, DataType.BOOLEAN):
                UnivariateAnalysis.top_most_frequent(df[col])
                UnivariateAnalysis.bar_plot(df[col])
            elif data_type == DataType.UNIQUE:
                display(Markdown("Each value in the column is unique."))
            elif data_type == DataType.MISSING:
                display(Markdown("The column contains only missing values."))
            else:
                UnivariateAnalysis.numeric_statistics(df[col])
                UnivariateAnalysis.histogram(df[col])

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook.
        """
        if self.verbosity == Verbosity.LOW:
            return [
                "from edvart.report_sections.univariate_analysis import UnivariateAnalysis\n"
                "univariate_analysis = UnivariateAnalysis.univariate_analysis"
            ]
        if self.verbosity == Verbosity.MEDIUM:
            return [
                "from edvart.report_sections.univariate_analysis import UnivariateAnalysis\n"
                "top_most_frequent = UnivariateAnalysis.top_most_frequent\n"
                "bar_plot = UnivariateAnalysis.bar_plot\n"
                "numeric_statistics = UnivariateAnalysis.numeric_statistics\n"
                "histogram = UnivariateAnalysis.histogram",
                "from edvart import utils",
            ]
        # Verbosity.HIGH
        return [
            "from edvart import utils",
            "from IPython.display import display",
            "from IPython.display import HTML",
            "import matplotlib.pyplot as plt",
            "%matplotlib inline",
            "import seaborn as sns",
        ]

    def add_cells(self, cells: List[Dict[str, Any]]) -> None:
        """Adds univariate analysis cells to the list of cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        """
        if self.columns is not None:
            self.df = self.df[self.columns]

        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)

        if self.verbosity == Verbosity.HIGH:
            # Add cell with default stats dictionaries
            default_stats_dicts = nbfv4.new_code_cell(
                "# Default statistics dictionaries"
                + 2 * "\n"
                + get_code(UnivariateAnalysis.default_descriptive_statistics)
                + 2 * "\n"
                + get_code(UnivariateAnalysis.default_quantile_statistics)
            )
            cells.append(default_stats_dicts)
            # Add cell with function definitions
            func_defs = nbfv4.new_code_cell(
                get_code(UnivariateAnalysis.top_most_frequent)
                + 2 * "\n"
                + get_code(UnivariateAnalysis.bar_plot)
                + 2 * "\n"
                + (
                    get_code(UnivariateAnalysis.numeric_statistics).replace(
                        "UnivariateAnalysis.", ""
                    )
                )
                + 2 * "\n"
                + get_code(UnivariateAnalysis.histogram)
            )
            cells.append(func_defs)
            # Add cell with dataframe rendering
            frame_rendering = nbfv4.new_code_cell(
                get_code(format_number)
                + 2 * "\n"
                + get_code(dict_to_html)
                + 2 * "\n"
                + get_code(add_html_heading)
                + 2 * "\n"
                + get_code(subcells_html)
            )
            cells.append(frame_rendering)

        if self.verbosity == Verbosity.LOW:
            if self.columns is None:
                code = "univariate_analysis(df=df)"
            else:
                code = f"univariate_analysis(df=df, columns={self.columns})"
            code_cell = nbfv4.new_code_cell(code)
            cells.append(code_cell)
        else:
            for col in self.df.columns:
                data_type = infer_data_type(self.df[col])
                data_type_name = str(data_type)
                column_header = nbfv4.new_markdown_cell(f"## *{col} - {data_type_name}*")
                cells.append(column_header)
                if data_type in (DataType.CATEGORICAL, DataType.BOOLEAN):
                    cell = nbfv4.new_code_cell(
                        code_dedent(
                            f"""
                            top_most_frequent(df['{col}'])
                            bar_plot(df['{col}'])"""
                        )
                    )
                elif data_type == DataType.UNIQUE:
                    cell = nbfv4.new_markdown_cell("Each value in the column is unique.")
                elif data_type == DataType.MISSING:
                    cell = nbfv4.new_markdown_cell("The column contains only missing values.")
                else:
                    if self.verbosity == Verbosity.MEDIUM:
                        cell = nbfv4.new_code_cell(
                            code_dedent(
                                f"""
                                numeric_statistics(df['{col}'])
                                histogram(df['{col}'])"""
                            )
                        )
                    else:
                        cell = nbfv4.new_code_cell(
                            code_dedent(
                                f"""
                                numeric_statistics(
                                    df['{col}'],
                                    descriptive_stats=default_descriptive_statistics(),
                                    quantile_stats=default_quantile_statistics()
                                )
                                histogram(df['{col}'])"""
                            )
                        )
                cells.append(cell)

    def show(self, df: pd.DataFrame) -> None:
        """Generates univariate analysis cell output in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        if self.columns is not None:
            df = df[self.columns]

        display(Markdown(self.get_title(section_level=1)))
        for col in df.columns:
            data_type = infer_data_type(df[col])
            data_type_name = str(data_type)
            display(Markdown(f"## *{col} - {data_type_name}*"))
            if data_type in (DataType.CATEGORICAL, DataType.BOOLEAN):
                UnivariateAnalysis.top_most_frequent(df[col])
                UnivariateAnalysis.bar_plot(df[col])
            elif data_type == DataType.UNIQUE:
                display(Markdown("Each value in the column is unique."))
            elif data_type == DataType.MISSING:
                display(Markdown("The column contains only missing values."))
            else:
                UnivariateAnalysis.numeric_statistics(df[col])
                UnivariateAnalysis.histogram(df[col])
            display(Markdown("---"))



"""Pandas formatting package."""

from typing import Any, Dict, List, Union

import pandas as pd
from IPython.display import display
from pandas.io.formats.style import Styler


def series_to_frame(series: pd.Series, index_name: str, column_name: str) -> pd.DataFrame:
    """Converts a pandas.Series to a pandas.DataFrame by putting the series index into a separate
    column.

    Parameters
    ---
    series : pd.Series
        Input series
    index_name : str
        Name of the new column into which the series index will be put
    column_name : str
        Name of the series values column

    Returns
    ---
    pd.DataFrame
        Dataframe with two columns index_name and column_name with values of series.index and
        series.values respectively
    """
    return series.rename_axis(index=index_name).to_frame(name=column_name).reset_index()


def render_dictionary(dictionary: Dict["str", Any]) -> None:
    """
    Converts a dictionary to a dataframe and renders that dataframe in the report notebook.

    Parameters
    ---
    dictionary: Dict['str', Any]
        Dictionary to be rendered
    """
    dictionary = {key: str(value) for key, value in dictionary.items()}

    display(pd.DataFrame.from_dict(dictionary, orient="index", columns=[""]))


def dict_to_html(dictionary: Dict["str", Any]) -> str:
    """Converts a dictionary to a dataframe in HTML string form.

    Parameters
    ----------
    dictionary : Dict['str', Any]
        DictDictionary to be converted
    """
    dictionary = {key: str(value) for key, value in dictionary.items()}

    return pd.DataFrame.from_dict(dictionary, orient="index", columns=[""]).to_html()


def add_html_heading(html: str, heading: str, heading_level: int = 2) -> str:
    """
    Adds a heading to an HTML string with the specified text and heading level.

    Parameters
    ---
    html : str
        HTML string to which to add heading
    heading : str
        Text of the heading
    heading_level : int
        Level of the heading

    Returns
    ---
    str
        HTML string with heading added
    """
    return f"<h{heading_level}> {heading} </h{heading_level}>" + html


def subcells_html(elements: List[List[str]]) -> str:
    """
    Returns HTML table in string format according to the elements matrix.

    Parameters
    ---
    elements : List[List[str]]
        Elements which should be rendered in table cells, outer list represents rows,
        inner list represents columns, elements themselves should be HTML strings

    Returns
    ---
    str
        Table in HTML string ready to be rendered for example by IPython.display.display_html
    """
    default_cell_css = """
        "
        text-align: left;
        vertical-align: text-top;
        align-content: center;
        "
    """

    i = 0
    html = '<table width="100%">'
    for row in elements:
        html += "<tr>"
        for element in row:
            if i == 0:
                html += f"<td style={default_cell_css}> {element} </td>"
                i = 1
            else:
                html += f"<td style={default_cell_css}> {element} </td>"
                i = 0
        html += "</tr>"
    html += "</table>"

    return html


def format_number(
    number: Union[int, float], decimal_places: int = 2, thousand_separator: str = ""
) -> str:
    """
    Formats a number by truncating decimal places (if it is a float) and optionally adds thousand
    separators.

    Parameters
    ---
    number : Union[int, float]
        Number to be converted to string
    decimal_places : int
        Number of decimal places in case of float
    thousand_separator : str
        Character or string with which thousands should be separated

    Returns
    ---
    str
        Formatted number in a string representation
    """
    if float(number).is_integer():
        formatted = f"{number:,}"
    else:
        formatted = f"{number:,.0{decimal_places}f}"
    return formatted.replace(",", thousand_separator)


def hide_index(df: pd.DataFrame) -> Styler:
    """
    Hides the index of a DataFrame.

    The method hide_index is deprecated since 1.4.0, but the method the replaces it
    is only available since 1.4.0, therefore we need to conditionally call one or the other
    to support both versions and avoid warnings (otherwise we would need to require
    python ^3.8, which is required by pandas 1.4.0.)

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame where the index should be hidden.

    Returns
    -------
    Styler
        Styler object with the index hidden.
    """
    if pd.__version__ < "1.4.0":
        return df.style.hide_index()
    return df.style.hide(axis="index")



"""Utils package."""

import os
from contextlib import contextmanager
from typing import Any, Dict, Iterable, Iterator, Literal, Optional, Tuple, Union

import pandas as pd
import statsmodels.api as sm


def top_frequent_values(series: pd.Series, n_top: int = 10) -> Dict[Any, float]:
    """
    Counts top n most frequent values in series along with other value counts and NULL value counts.

    Parameters
    ---
    series: pd.Series
        Input series of data for which frequencies will be calculated
    n_top: int
        Number of values for which actual frequencies will be counted, other values will be grouped
        into 'Other' category

    Returns
    -------
    result_dict: Dict
        Dictionary with the mapping {'value': 'frequency (relative frequency)'}
    """
    # Calculate frequencies
    counts = series.value_counts()
    nan_count = series.isna().sum()
    result_dict = {
        **(counts[:n_top].to_dict()),
        "Other values count": counts[n_top:].sum(),
        "Null": nan_count,
    }

    # Add relative frequencies
    for key, _value in result_dict.items():
        result_dict[key] = f"{result_dict[key]:,} ({100 * result_dict[key] / len(series):.02f} %)"

    return result_dict


def reindex_to_datetime(
    df: pd.DataFrame,
    datetime_column: str,
    keep_index: Optional[str] = None,
    unit: str = "ns",
    origin: Union[str, pd.Timestamp] = "unix",
    sort: bool = True,
) -> pd.DataFrame:
    """Reindex a given DataFrame to be indexed by a pd.DateTimeIndex.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to reindex.
    datetime_column : str
        Which column containing datetimes to index by.
    keep_index : str, optional
        Name of column to store the original index. The original index will be discarded
        by default.
    unit : str (default = "ns")
        Numeric values would be parsed as number of units from origin.
    origin : Union[str, pd.Timestamp] (default = "unix")
        Define the reference date. Numeric values would be parsed as number of units
        (defined by unit) since this reference date. By default unix epoch 1970-01-01.
    sort : bool (default = True)
        Whether to sort according to the index.

    Returns
    -------
    pd.DataFrame
        Reindexed df.
    """
    df = df.copy()
    new_index = pd.to_datetime(
        df[datetime_column], unit=unit, origin=origin, infer_datetime_format=True
    )
    if keep_index is not None:
        df[keep_index] = df.index
    df = df.drop(datetime_column, axis="columns")
    df.index = pd.DatetimeIndex(new_index)
    if sort:
        df = df.sort_index()

    return df


def reindex_to_period(
    df: pd.DataFrame,
    period_column: str,
    freq: Optional[str],
    keep_index: Optional[str] = None,
    sort: bool = True,
) -> pd.DataFrame:
    """Reindex a given DataFrame to be indexed by a pd.PeriodIndex.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to reindex.
    period_column : str
        Which column containing periods to index by.
    freq : Union[str, pd.Offset]
        One of pandas’ offset strings or an Offset object. Will be inferred by default.
    keep_index : str, optional
        Name of column to store the original index. The original index will be discarded
        by default.
    sort : bool (default = True)
        Whether to sort according to the index.

    Returns
    -------
    pd.DataFrame
        Reindexed df.
    """
    df = df.copy()
    new_index = df[period_column].to_period(freq=freq)
    if keep_index is not None:
        df[keep_index] = df.index
    df = df.drop(period_column, axis="columns")
    df.index = pd.PeriodIndex(new_index)
    if sort:
        df = df.sort_index()

    return df


def discrete_colorscale(n, saturation=0.5, lightness=0.5) -> Iterable[Tuple[float, str]]:
    """
    Generate a colorscale of n discrete colors.

    Colours are equally spaced around the complete HSL wheel with constant saturation and lightness.

    Returns
    -------
    Iterable[Tuple[float, str]]
        An iterable of 2n tuples, where each tuple contains a value between 0 and 1
        (the values are equally and each value appears twice) and an HSL string containing
        an HSL color string with hue corresponding to the value.
    """
    for i in range(n):
        color = f"hsl({(i / n) * 360 :.2f}, {saturation * 100 :.2f}%, {lightness * 100 :.2f}%)"
        yield (i / n, color)
        yield ((i + 1) / n, color)


#######################
# Statistic functions #
#######################


def num_unique_values(series: pd.Series) -> int:
    """
    Return number of unique values.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    int
    """
    return series.nunique()


def median_absolute_deviation(series: pd.Series) -> float:
    """
    Return median absolute deviation.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return median((series - series.mean()).abs())


def coefficient_of_variation(series: pd.Series) -> float:
    """
    Return coefficient of variation.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.std() / series.mean()


def minimum(series: pd.Series) -> float:
    """
    Return minimum.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.min()


def maximum(series: pd.Series) -> float:
    """
    Return maximum.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.max()


def quartile1(series: pd.Series) -> float:
    """
    Return first quartile.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.quantile(0.25)


def quartile3(series: pd.Series) -> float:
    """
    Return third quartile.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.quantile(0.75)


def mean(series: pd.Series) -> float:
    """
    Return mean.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.mean()


def median(series: pd.Series) -> float:
    """
    Return median.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return series.median()


def iqr(series: pd.Series) -> float:
    """
    Return inter quartile range.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.quantile(0.75) - series.quantile(0.25)


def value_range(series: pd.Series) -> float:
    """
    Return value range.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.max() - series.min()


def mode(series: pd.Series) -> float:
    """
    Return mode.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
        The most frequent value. `float('nan')` if the series contains only null values.
    """
    most_frequent = series.mode(dropna=True)
    if len(most_frequent) == 0:
        return float("nan")
    return most_frequent[0]


def std(series: pd.Series) -> float:
    """
    Return standard deviation.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.std()


def mad(series: pd.Series) -> Any:
    """
    Return mean absolute deviation.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return (series - series.mean()).abs().mean()


def kurtosis(series: pd.Series) -> Any:
    """
    Return kurtosis.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.kurtosis()


def skewness(series: pd.Series) -> Any:
    """
    Return skewness.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.skew()


def sum_(series: pd.Series) -> float:
    """
    Return sum.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.sum(min_count=1)


def pearson(df: pd.DataFrame) -> pd.DataFrame:
    """
    Return pearson correlation coefficient.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame on which the stat should be calculated.

    Returns
    -------
    pd.DataFrame
    """
    return _corr(df, "pearson")


def spearman(df: pd.DataFrame) -> pd.DataFrame:
    """
    Return spearman correlation coefficient.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame on which the stat should be calculated.

    Returns
    -------
    pd.DataFrame
    """
    return _corr(df, "spearman")


def kendall(df: pd.DataFrame) -> pd.DataFrame:
    """
    Return kendall correlation coefficient.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame on which the stat should be calculated.

    Returns
    -------
    pd.DataFrame
    """
    return _corr(df, "kendall")


def _corr(df: pd.DataFrame, method: Literal["pearson", "kendall", "spearman"]) -> pd.DataFrame:
    return df.corr(method=method, numeric_only=True)


def contingency_table(df: pd.DataFrame) -> pd.DataFrame:
    """
    Return contingency table.

    Parameters
    ----------
    df: pd.DataFrame

    Returns
    -------
    pd.DataFrame
    """
    table = sm.stats.Table.from_data(df)
    return table.table_orig.astype(int)


@contextmanager
def env_var(name: str, value: str) -> Iterator[None]:
    """
    Set an environment variable for the duration of the context.

    Parameters
    ----------
    name : str
        Name of the environment variable.
    value : str
        Value of the environment variable.
    """
    original_env = os.environ.copy()
    os.environ[name] = value
    try:
        yield
    finally:
        os.environ = original_env



"""Group analysis module."""

from typing import Any, Callable, Dict, List, Optional, Union

import colorlover as cl
import nbformat.v4 as nbfv4
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from IPython.display import Markdown, display
from matplotlib import cm
from matplotlib.colors import Normalize, to_hex
from plotly.subplots import make_subplots

from edvart import utils
from edvart.data_types import DataType, infer_data_type
from edvart.report_sections.code_string_formatting import code_dedent, get_code, total_dedent
from edvart.report_sections.section_base import Section, Verbosity


class GroupAnalysis(Section):
    """Generate the group analysis section of the report.

    Parameters
    ----------
    df : pd.DataFrame
        Data for which to perform analysis.
    groupby : Union[str, List[str]]
        Name of column or list of columns names to group by.
    verbosity : Verbosity (default = Verbosity.LOW)
        Generated code verbosity global to the Group analysis sections.
        If subsection verbosities are None, then they will be overridden by this parameter.
    columns : List[str], optional
        Columns on which to do group analysis. All columns are used by default.
    show_within_group_statistics : bool (default = True)
        Whether to show per-group statistics.
    show_group_missing_values : bool (default = True)
        Whether to show per-group missing values.
    show_group_distribution_plots : bool (default = True)
        Whether to show per-group distribution plots.

    Raises
    ------
    ValueError
        If groupby columns are not a subset of the columns of the input DataFrame df.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        groupby: Union[str, List[str]],
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        show_within_group_statistics: bool = True,
        show_group_missing_values: bool = True,
        show_group_distribution_plots: bool = True,
    ):
        if isinstance(groupby, str):
            groupby = [groupby]
        if not set(groupby) <= set(df.columns):
            raise ValueError("Grouping by a column which is not in columns of df.")
        self.columns = columns
        self.df = df
        self.groupby = groupby
        self.show_statistics = show_within_group_statistics
        self.show_missing_vals = show_group_missing_values
        self.show_dist = show_group_distribution_plots

        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Group Analysis"

    @staticmethod
    def default_group_descriptive_stats() -> Dict[str, Callable[[pd.Series], float]]:
        """Descriptive statistic functions.

        Returns
        -------
        Dict[str, Callable[[pd.Series], float]]
            A dictionary of statistic function names and functions.
        """
        return {
            "# Unique values": utils.num_unique_values,
            "Sum": utils.sum_,
            "Mode": utils.mode,
            "Mean": utils.mean,
            "Std": utils.std,
            "Mean abs dev": utils.mad,
            "Median abs dev": utils.median_absolute_deviation,
            "Relative std": utils.coefficient_of_variation,
            "Kurtosis": utils.kurtosis,
            "Skewness": utils.skewness,
        }

    @staticmethod
    def default_group_quantile_stats() -> Dict[str, Callable[[pd.Series], float]]:
        """Quantile statistic functions.

        Returns
        -------
        Dict[str, Callable[[pd.Series], float]]
            A dictionary of statistic function names and functions.
        """
        return {
            "Min": utils.minimum,
            "Q1": utils.quartile1,
            "Median": utils.median,
            "Q3": utils.quartile3,
            "Max": utils.maximum,
            "Range": utils.value_range,
            "IQR": utils.iqr,
        }

    @staticmethod
    def within_group_descriptive_stats(
        df: pd.DataFrame, groupby: List[str], column: str, round_decimals: int = 2
    ):
        """Display within-group descriptive statistics for a column.

        Parameters
        ----------
        df : pd.DataFrame
            Data to display statistics for.
        groupby : List[str]
            List of column names to group data by.
        column : str
            Which column to display statistics for.
        round_decimals : int (default = 2)
            Number of decimals to round displayed results to.
        """
        GroupAnalysis.within_group_stats(
            df=df,
            groupby=groupby,
            column=column,
            stats=GroupAnalysis.default_group_descriptive_stats(),
            round_decimals=round_decimals,
        )

    @staticmethod
    def within_group_quantile_stats(
        df: pd.DataFrame, groupby: List[str], column: str, round_decimals: int = 2
    ) -> None:
        """Display within-group quantile statistics for a column.

        Parameters
        ----------
        df : pd.DataFrame
            Data to display statistics for.
        groupby : List[str]
            List of column names to group data by.
        column : str
            Which column to display statistics for.
        round_decimals : int (default = 2)
            Number of decimals to round displayed results to.
        """
        GroupAnalysis.within_group_stats(
            df=df,
            groupby=groupby,
            column=column,
            stats=GroupAnalysis.default_group_quantile_stats(),
            round_decimals=round_decimals,
        )

    @staticmethod
    def within_group_stats(
        df: pd.DataFrame,
        groupby: List[str],
        column: str,
        stats: Dict[str, Callable[[pd.Series], float]] = None,
        round_decimals: int = 2,
    ) -> None:
        """Display withing group statistics for a column of df grouped by one or other more columns.

        Parameters
        ----------
        df : pd.DataFrame
            Data to display statistics for.
        groupby : List[str]
            List of column names to group by.
        column : str
            Name of column to display statistics for.
        stats : Dict[str, Callable[[pd.Series], float]], optional
            A dictionary of statistic function names and functions.
            If None, default_group_quantile_stats() and default_group_descriptive_stats() will be
            used.
        round_decimals : int (default = 2)
            Number of decimals to round displayed results to.
        """
        if stats is None:
            GroupAnalysis.within_group_descriptive_stats(df, groupby, column, round_decimals)
            GroupAnalysis.within_group_quantile_stats(df, groupby, column, round_decimals)
            return

        df_grouped = df.groupby(groupby)[column]
        group_stats = []
        for name, func in stats.items():
            group_stats.append(df_grouped.apply(func).rename(name))
        stats_table = pd.concat(group_stats, axis=1)
        stats_table = stats_table.round(decimals=round_decimals)
        display(stats_table)

    # pylint: disable=too-many-locals
    @staticmethod
    def group_missing_values(
        df: pd.DataFrame,
        groupby: Union[str, List[str]],
        columns: Optional[List[str]] = None,
        round_decimals: int = 2,
        heatmap: bool = True,
        foreground_colormap: str = "bone",
        background_colormap: str = "OrRd",
        sort: bool = True,
        sort_by: Optional[List[str]] = None,
        ascending: bool = False,
    ) -> None:
        """Display per-group number and percentage of missing values in each column.

        Parameters
        ----------
        df : pd.DataFrame
            Data to display missing values for.
        groupby : str or List[str]
            Name of column or list of columns names to group by.
        columns : List[str], optional
            Subset of columns to analyze. All columns except those for grouping are used by default.
        round_decimals : int (default = 2)
            Number of decimals to round displayed results to.
        heatmap : bool (default = True)
            Whether to color missing value percentage cells according to the corresponding value.
        foreground_colormap : str (default = "bone")
            Color map of the foreground.
        background_colormap : str (default = "OrRd)
            Color map of the background.
        sort : bool (default = True)
            Whether to sort the results.
        sort_by : List[str], optional
            List of column names to sort the results by. Sort by all column by default.
        ascending : bool (default = False)
            If True, sort in ascending order, otherwise sort in descending order.

        Raises
        ------
        ValueError
            If groupby columns are not a subset of the columns of the input DataFrame df.
        """
        if isinstance(groupby, str):
            groupby = [groupby]
        if not set(groupby) <= set(df.columns):
            raise ValueError("Grouping by a column which is not in columns of df.")
        if columns is None:
            columns = [col for col in df.columns if col not in groupby]
        df_grouped = df.groupby(groupby)[columns]

        # Calculate number of samples in each group
        sizes = df_grouped.size().rename("Group Size")

        # Calculate missing values percentage of each column for each group
        missing = df_grouped.apply(lambda g: g.isna().sum(axis=0))
        missing = missing.divide(sizes, axis=0) * 100  # `/` divides by axis=-1
        missing = missing.fillna(value=0)
        missing = missing.round(decimals=round_decimals)

        if missing.sum().sum() == 0:
            print("There are no missing values")
            return

        # Concatenate group sizes and missing value percentages
        final_table = pd.concat([sizes, missing], axis=1)

        # Sort columns to better identify groups with missing data
        all_columns = [col for col in missing.columns if col not in groupby]
        if sort:
            if sort_by is None:
                sort_by = all_columns
            final_table = final_table.sort_values(by=sort_by, axis=0, ascending=ascending)

        # Drop columns with no missing data
        missing = final_table.sum(axis=0) != 0
        final_table = final_table.loc[:, missing]

        colored_columns = [col for col in final_table if col in all_columns]

        # Apply conditional formatting to each cell except group size column
        if heatmap:
            fg_cmap = cm.get_cmap(foreground_colormap)
            bg_cmap = cm.get_cmap(background_colormap)
            norm = Normalize(vmin=0, vmax=100)

            def color_cell(value):
                fg_hex = to_hex(fg_cmap(norm(value)))
                bg_hex = to_hex(bg_cmap(norm(value)))
                return f"""
                    color: {fg_hex};
                    background-color: {bg_hex};
                """

            render = final_table.style.applymap(
                func=color_cell, subset=pd.IndexSlice[:, colored_columns]
            ).format(formatter="{0:.2f} %", subset=pd.IndexSlice[:, colored_columns])
        else:
            render = final_table.style.format(
                formatter="{0:.2f} %", subset=pd.IndexSlice[:, colored_columns]
            )

        # Render table
        display(render)

    @staticmethod
    def group_barplot(
        df: pd.DataFrame,
        groupby: List[str],
        column: str,
        group_count_threshold: int = 20,
        conditional_probability: bool = True,
        xaxis_tickangle: float = 0,
        alpha: float = 0.5,
    ):
        """Display a per-group barplot for a column.

        Parameters
        ----------
        df : pd.DataFrame
            Data to analyze.
        groupby : List[str]
            List of column names to group by.
        column : str
            Which column to analyze.
        group_count_threshold : int (default = 20)
            Maximum number of unique values in column to plot. If the number of unique values
            is higher, a warning will be issued and plot will not be shown.
        conditional_probability : bool (default = True)
            If True, conditional probability contitioned on group will be displayed,
            otherwise conditional frequency will be displayed.
        xaxis_tickangle : float (default = 0)
            Rotation angle of ticks on the x axis.
        alpha : float (default = 0.5)
            Opacity of bars in the plot.
        """
        num_cat = df[column].nunique()
        if num_cat > group_count_threshold:
            display(
                Markdown(
                    f"Number of unique values ({num_cat}) greater than threshold."
                    " Not plotting distribution"
                )
            )
            return

        pivot = df.pivot_table(index=groupby, columns=column, aggfunc="size", fill_value=0)

        if conditional_probability:
            pivot = pivot.divide(pivot.sum(axis=1), axis=0)  # `/` divides by axis=-1
            pivot = pivot.fillna(value=0)

        # Choose color palette
        colors = cl.scales["9"]["qual"]["Set1"]
        color_idx = 0

        fig = go.Figure()
        for idx, row in pivot.iterrows():
            if hasattr(idx, "__len__") and not isinstance(idx, str):
                group_name = "_".join([str(i) for i in idx])
            else:
                group_name = idx
            color = colors[color_idx % len(colors)]
            color_idx += 1
            fig.add_trace(
                go.Bar(
                    x=pivot.columns,
                    y=row,
                    name=group_name,
                    opacity=alpha,
                    marker_color=color,
                )
            )

        groupby_repr = ", ".join(groupby)
        if conditional_probability:
            yaxis_title = f"P({column} | {groupby_repr})"
        else:
            yaxis_title = f"Freq({column} | {groupby_repr})"

        fig.update_layout(
            barmode="group",
            xaxis_tickangle=xaxis_tickangle,
            xaxis_title=column,
            yaxis_title=yaxis_title,
        )
        fig.show()

    @staticmethod
    def overlaid_histograms(
        df: pd.DataFrame,
        groupby: List[str],
        column: str,
        bins: Optional[int] = None,
        density: bool = True,
        alpha: float = 0.5,
    ):
        """Show per-group distribution histograms in a single plot overlaid over each other.

        Parameters
        ----------
        df : pd.DataFrame
            Data to analyze.
        groupby : List[str]
            List of column names to group by.
        column : str
            Name of column to analyze.
        bins : int, optional
            Number of bins in the histogram. If None, number of bin will be inferred using
            Modified Freedman-Diaconis bin number inference.
        density : bool (default = True)
            If True, histograms will be normalized to display density.
        alpha : float
            Opacity of individual histograms.
        """
        # Modified Freedman-Diaconis bin number inference if bins is None
        if bins is None:
            iqr = df[column].quantile(0.75) - df[column].quantile(0.25)
            bin_width = 1 / np.cbrt(len(df)) * iqr
        else:
            bin_width = (df[column].max() - df[column].min()) / bins
        bin_config = {
            "start": df[column].min(),
            "end": df[column].max(),
            "size": bin_width,
        }

        # Choose color palette
        colors = cl.scales["9"]["qual"]["Set1"]
        color_idx = 0

        # Distribution plot
        fig = make_subplots(
            rows=2,
            cols=1,
            shared_xaxes=True,
            row_heights=[0.3, 0.7],
            vertical_spacing=0.02,
        )

        for name, group in df.groupby(
            by=(groupby[0] if isinstance(groupby, list) and len(groupby) == 1 else groupby)
        ):
            if hasattr(name, "__len__") and not isinstance(name, str):
                group_name = "_".join([str(i) for i in name])
            else:
                group_name = name
            color = colors[color_idx % len(colors)]
            color_idx += 1
            # Add to boxplot
            fig.add_trace(
                go.Box(
                    x=group[column],
                    name=group_name,
                    legendgroup=group_name,
                    showlegend=False,
                    marker_color=color,
                ),
                row=1,
                col=1,
            )
            # Add to histogram
            fig.add_trace(
                go.Histogram(
                    x=group[column],
                    name=group_name,
                    legendgroup=group_name,
                    xbins=bin_config,
                    histnorm="density" if density else "",
                    marker_color=color,
                    opacity=alpha,
                ),
                row=2,
                col=1,
            )
        fig.update_layout(barmode="overlay")
        fig.update_xaxes(title_text=column, row=2, col=1)
        yaxis_title = "Density" if density else "Frequency"
        fig.update_yaxes(title_text=yaxis_title, row=2, col=1)
        fig.show()

    @staticmethod
    def group_analysis(
        df: pd.DataFrame,
        groupby: Union[str, List[str]],
        columns: Optional[List[str]] = None,
        show_within_group_statistics: bool = True,
        show_group_missing_values: bool = True,
        show_distribution_plots: bool = True,
    ) -> None:
        """Generate group analysis for df.

        Parameters
        ----------
        df : pd.DataFrame
            Data to be analyzed.
        groupby : Union[str, List[str]]
            Name of column or list of columns names to group by.
        columns : List[str], optional
            Subset of columns to analyze. All columns except those used for grouping are used
            by default.
        show_within_group_statistics : bool (default = True)
            Whether to show per-group statistics.
        show_group_missing_values : bool (default = True)
            Whether to show per-group missing values.
        show_distribution_plots : bool (default = True)
            Whether to show per-group distribution plots.

        Raises
        ------
        ValueError
            If groupby columns are not a subset of the columns of the input DataFrame df.
        """
        if isinstance(groupby, str):
            groupby = [groupby]
        if not set(groupby) <= set(df.columns):
            raise ValueError("Grouping by a column which is not in columns of df.")
        if columns is None:
            columns = [col for col in df.columns if col not in groupby]

        if show_group_missing_values:
            display(Markdown("## Missing values for each group"))
            GroupAnalysis.group_missing_values(df, groupby)
        if not show_distribution_plots and not show_within_group_statistics:
            return
        for col in columns:
            display(Markdown("---"))
            display(Markdown(f"### *{col}*"))
            datatype = infer_data_type(df[col])
            if datatype == DataType.NUMERIC:
                if show_within_group_statistics:
                    GroupAnalysis.within_group_stats(df, groupby, col)
                if show_distribution_plots:
                    GroupAnalysis.overlaid_histograms(df, groupby, col)
            else:
                GroupAnalysis.group_barplot(df, groupby, col)

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"]
        """
        ga = "GroupAnalysis"  # pylint:disable=invalid-name
        if self.verbosity == Verbosity.LOW:
            return [
                f"from edvart.report_sections.group_analysis import {ga}\n"
                f"group_analysis = {ga}.group_analysis"
            ]
        if self.verbosity == Verbosity.MEDIUM:
            return [
                total_dedent(
                    f"""
                        from edvart.report_sections.group_analysis import {ga}
                        group_missing_values = {ga}.group_missing_values
                        within_group_stats = {ga}.within_group_stats
                        within_group_descriptive_stats = {ga}.within_group_descriptive_stats
                        within_group_quantile_stats = {ga}.within_group_quantile_stats
                        group_barplot = {ga}.group_barplot
                        overlaid_histograms = {ga}.overlaid_histograms
                        default_group_quantile_stats = {ga}.default_group_quantile_stats
                        default_group_descriptive_stats = {ga}.default_group_descriptive_stats
                    """
                )
            ]
        # verbosity HIGH
        return [
            "import colorlover as cl",
            "import matplotlib.pyplot as plt",
            "import matplotlib.cm as cm",
            "from matplotlib.colors import Normalize, to_hex",
            "import numpy as np",
            "import pandas as pd",
            "import plotly.graph_objects as go",
            "from edvart.data_types import infer_data_type, DataType",
            "from edvart import utils",
            "from typing import List, Dict, Optional, Callable",
            "from plotly.subplots import make_subplots",
        ]

    def _add_function_defs(self, cells: List[Dict[str, Any]]):
        """Add function definition cell to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        """
        code = (
            get_code(GroupAnalysis.default_group_quantile_stats)
            + "\n\n"
            + get_code(GroupAnalysis.default_group_descriptive_stats)
            + "\n\n"
            + (get_code(GroupAnalysis.within_group_descriptive_stats).replace("GroupAnalysis.", ""))
            + "\n\n"
            + (get_code(GroupAnalysis.within_group_quantile_stats).replace("GroupAnalysis.", ""))
            + "\n\n"
            + (get_code(GroupAnalysis.within_group_stats).replace("GroupAnalysis.", ""))
            + "\n\n"
            + get_code(GroupAnalysis.group_barplot)
            + "\n\n"
            + get_code(GroupAnalysis.overlaid_histograms)
            + "\n\n"
            + get_code(GroupAnalysis.group_missing_values)
        )

        cells.append(nbfv4.new_code_cell(code))

    def _add_cells_numeric_col(self, cells: List[Dict[str, Any]], column_name: str):
        """Add code cells for a numeric column at verbosity MEDIUM or HIGH.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        column_name : str
            Name of column for which to generate code.
        """
        code = ""
        if self.show_statistics:
            if self.verbosity == Verbosity.MEDIUM:
                code += (
                    f"within_group_stats(df=df, groupby={self.groupby}, column='{column_name}')\n"
                )
            else:
                code += code_dedent(
                    f"""
                            within_group_stats(
                                df=df,
                                groupby={self.groupby},
                                column='{column_name}',
                                stats=default_group_descriptive_stats()
                            )

                            within_group_stats(
                                df=df,
                                groupby={self.groupby},
                                column='{column_name}',
                                stats=default_group_quantile_stats()
                            )
                        """
                )

        if self.show_dist:
            code += f"overlaid_histograms(df=df, groupby={self.groupby}, column='{column_name}')"
        cells.append(nbfv4.new_code_cell(code))

    def add_cells(self, cells: List[Dict[str, Any]]):
        """Add cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)

        if self.verbosity == Verbosity.LOW:
            if self.columns is None:
                code = f"group_analysis(df=df, groupby={self.groupby})"
            else:
                code = f"group_analysis(df=df, groupby={self.groupby}, columns={self.columns})"
            code_cell = nbfv4.new_code_cell(code)
            cells.append(code_cell)
            return

        if self.verbosity == Verbosity.HIGH:
            self._add_function_defs(cells)

        if self.show_missing_vals:
            cells.append(nbfv4.new_markdown_cell("## Missing values for each group"))
            if self.columns is None:
                code = f"group_missing_values(df=df, groupby={self.groupby})"
            else:
                code = (
                    f"group_missing_values(df=df, columns={self.columns}, groupby={self.groupby})"
                )
            cells.append(nbfv4.new_code_cell(code))

        columns = self.columns if self.columns is not None else self.df.columns

        if not self.show_statistics and not self.show_dist:
            return
        for col in columns:
            if col in self.groupby:
                continue
            cells.append(nbfv4.new_markdown_cell(f"## *{col}*"))
            datatype = infer_data_type(self.df[col])
            if datatype == DataType.NUMERIC:
                self._add_cells_numeric_col(cells, col)
            else:
                code = f"group_barplot(df=df, groupby={self.groupby}, column='{col}')"
                cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output of this section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output.
        """
        columns = (
            self.columns
            if self.columns is not None
            else [col for col in df.columns if col not in self.groupby]
        )
        display(Markdown(self.get_title(section_level=1)))

        if self.show_missing_vals:
            display(Markdown("## Missing values for each group"))
            GroupAnalysis.group_missing_values(df, self.groupby)
        if not self.show_dist and not self.show_statistics:
            return
        for col in columns:
            if col in self.groupby:
                continue
            display(Markdown(f"## *{col}*"))
            datatype = infer_data_type(df[col])
            if datatype == DataType.NUMERIC:
                if self.show_statistics:
                    GroupAnalysis.within_group_stats(df, self.groupby, col)
                if self.show_dist:
                    GroupAnalysis.overlaid_histograms(df, self.groupby, col)
            else:
                GroupAnalysis.group_barplot(df, self.groupby, col)



"""Multivariate analysis package."""

from enum import IntEnum
from typing import Any, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import sklearn.decomposition
from IPython.display import Markdown, display
from sklearn.preprocessing import StandardScaler

from edvart.data_types import is_boolean, is_categorical, is_numeric
from edvart.plots import scatter_plot_2d
from edvart.report_sections.code_string_formatting import get_code, total_dedent
from edvart.report_sections.section_base import ReportSection, Section, Verbosity
from edvart.utils import discrete_colorscale

try:
    from edvart.report_sections.umap import UMAP
except ImportError:
    UMAP_AVAILABLE = False
else:
    UMAP_AVAILABLE = True


class MultivariateAnalysis(ReportSection):
    """Generates the Multivariate analysis section of the report.

    Contains an enum MultivariateAnalysisSubsection of possible subsections.

    Parameters
    ----------
    df : pd.DataFrame
        Data for which to perform analysis.
    subsections : List[MultivariateAnalysisSubsection], optional
        List of subsections to include.
        All subsection in MultivariateAnalysisSubsection are included by default.
    verbosity : Verbosity
        Generated code verbosity global to the Multivariate sections.
        If subsection verbosities are None, then they will be overridden by this parameter.
    columns : List[str], optional
        Columns on which to do multivariate analysis.
        All columns of df will be used by default.
    verbosity_pca : Verbosity, optional
        Principal component analysis subsection code verbosity.
    verbosity_umap : Verbosity, optional
        UMAP subsection code verbosity.
    verbosity_parallel_coordinates : Verbosity, optional
        Parallel coordinates subsection code verbosity.
    verbosity_parallel_categories : Verbosity, optional
        Parallel categories subsection code verbosity.
    color_col : str, optional
        Name of the column according to which to color points in the sections.
        Both numerical and categorical columns are supported.
    """

    # pylint:disable=invalid-name
    class MultivariateAnalysisSubsection(IntEnum):
        """Enum of all implemented multivariate analysis subsections."""

        PCA = 0
        if UMAP_AVAILABLE:
            UMAP = 1
        ParallelCoordinates = 2
        ParallelCategories = 3

        def __str__(self):
            return self.name

    def __init__(
        self,
        df: pd.DataFrame,
        subsections: Optional[List[MultivariateAnalysisSubsection]] = None,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        verbosity_pca: Optional[Verbosity] = None,
        verbosity_umap: Optional[Verbosity] = None,
        verbosity_parallel_coordinates: Optional[Verbosity] = None,
        verbosity_parallel_categories: Optional[Verbosity] = None,
        color_col: Optional[str] = None,
    ):
        verbosity_pca = verbosity_pca or verbosity
        verbosity_umap = verbosity_umap or verbosity
        verbosity_parallel_coordinates = (
            verbosity_parallel_coordinates
            if verbosity_parallel_coordinates is not None
            else verbosity
        )
        verbosity_parallel_categories = (
            verbosity_parallel_categories
            if verbosity_parallel_categories is not None
            else verbosity
        )

        subsec = MultivariateAnalysis.MultivariateAnalysisSubsection
        verbosities = {
            subsec.PCA: verbosity_pca,
            subsec.ParallelCoordinates: verbosity_parallel_coordinates,
            subsec.ParallelCategories: verbosity_parallel_categories,
        }
        if UMAP_AVAILABLE:
            verbosities[subsec.UMAP] = verbosity_umap

        # By default use all subsections
        if subsections is None:
            subsections_all = list(subsec)
        else:
            subsections_all = subsections

        # Store subsections with LOW verbosity
        self.subsections_low_verbosity = [
            sub for sub in subsections_all if verbosities[sub] == Verbosity.LOW
        ]

        if len(self.subsections_low_verbosity) == len(subsections_all) and subsections is None:
            self.subsections_low_verbosity = None

        enum_to_implementation = {
            subsec.PCA: PCA(df, verbosity_pca, columns, color_col=color_col),
            subsec.ParallelCoordinates: ParallelCoordinates(
                df, verbosity_parallel_coordinates, columns, color_col=color_col
            ),
            subsec.ParallelCategories: ParallelCategories(
                df, verbosity_parallel_categories, columns, color_col=color_col
            ),
        }
        if UMAP_AVAILABLE:
            enum_to_implementation[subsec.UMAP] = UMAP(
                df, verbosity_umap, columns, color_col=color_col
            )

        subsections_implementations = [enum_to_implementation[sub] for sub in subsections_all]

        self.color_col = color_col
        super().__init__(subsections_implementations, verbosity, columns)

    @property
    def name(self) -> str:
        return "Multivariate analysis"

    @staticmethod
    def multivariate_analysis(
        df: pd.DataFrame,
        subsections: Optional[List[MultivariateAnalysisSubsection]] = None,
        columns: Optional[List[str]] = None,
        color_col: Optional[str] = None,
    ) -> None:
        """Generates multivariate analysis for df.

        Parameters
        ----------
        df : pd.DataFrame
            Data to be analyzed
        subsections : List[MultivariateAnalysisSubsection], optional
            Subsections to include in the analysis. All subsections are included by default.
        columns : List[str], optional
            Subset of columns of df to consider in multivariate analysis.
            All numeric columns are used by default.
        color_col : str, optional
            Name of the column according to which to color points in the sections.
            Both numberic and categorical columns are supported.
        """
        if columns is not None:
            df = df[columns]

        multivariate_analysis = MultivariateAnalysis(
            df=df,
            subsections=subsections,
            verbosity=Verbosity.LOW,
            columns=columns,
            color_col=color_col,
        )

        for sub in multivariate_analysis.subsections:
            sub.show(df)

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np']
        """
        if self.verbosity == Verbosity.LOW:
            imports = {
                "from edvart.report_sections.multivariate_analysis import MultivariateAnalysis\n"
                "multivariate_analysis = MultivariateAnalysis.multivariate_analysis"
            }
            for subsec in self.subsections:
                if subsec.verbosity > Verbosity.LOW:
                    imports.update(subsec.required_imports())

            return list(imports)
        return super().required_imports()

    def add_cells(self, cells: List[Dict[str, Any]]) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)
        if self.verbosity == Verbosity.LOW:
            code = "multivariate_analysis(df=df"
            if self.subsections_low_verbosity is not None:
                arg_subsections_names = [
                    f"MultivariateAnalysis.MultivariateAnalysisSubsection.{str(sub)}"
                    for sub in self.subsections_low_verbosity
                ]
                code += f", subsections={arg_subsections_names}".replace("'", "")
            if self.columns is not None:
                code += f", columns={self.columns}"
            if self.color_col is not None:
                code += f", color_col='{self.color_col}'"
            code += ")"
            cells.append(nbfv4.new_code_cell(code))
            for sub in self.subsections:
                if sub.verbosity > Verbosity.LOW:
                    sub.add_cells(cells)
        else:
            super().add_cells(cells)

    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output of this section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output.
        """
        display(Markdown(self.get_title(section_level=1)))
        super().show(df)


class PCA(Section):
    """Generates the Principal component analysis subsection.

    Parameters
    ----------
    df : pd.DataFrame
        Data on which to perform PCA.
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns on which to perform PCA. Only numeric columns can be used.
        All numeric columns of df are used by default.
    color_col : str, optional
        Name of column according to values of which to color points in the first vs second component
        plot. Can be both numeric and categorical. By default, all points have the same color.
    standardize : bool (default = True)
        Whether to standardize the data to zero mean and unit variance before applying PCA.
    interactive : bool (default = True)
        Whether to plot the first vs second principal component as an interactive plot.
        The interactive plot also shows labels for each point on hover.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        color_col: Optional[str] = None,
        standardize: bool = True,
        interactive: bool = True,
    ):
        # By default use only numeric columns
        if columns is None:
            columns = [col for col in df.columns if is_numeric(df[col])]
            # If all columns are numeric we don't want to list them all in the generated call
            # Setting columns to None will result in the columns argumented not being included
            # instead of showing a potentially long list of all columns
            # in the generated call, therefore the default (all columns) will be used
            if len(columns) == len(df.columns):
                columns = None
        else:
            for col in columns:
                if not is_numeric(df[col]):
                    raise ValueError(
                        f"Cannot use non-numeric column {col} of dtype {df[col].dtype} in PCA"
                    )

        self.color_col = color_col
        self.interactive = interactive
        self.standardize = standardize
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Principal Component Analysis"

    @staticmethod
    def pca_first_vs_second(
        df: pd.DataFrame,
        columns: Optional[List[str]] = None,
        color_col: Optional[str] = None,
        interactive: bool = True,
        standardize: bool = True,
        figsize: Tuple[float, float] = (12, 12),
        opacity: float = 0.8,
    ) -> None:
        """Plot a 2D scatter of first vs second PCA components.

        Parameters
        ----------
        df : pd.DataFrame
            Data to perform PCA on.
        columns : List[str], optional
            Which columns to perform PCA on. All columns will be used by default.
        color_col : str, optional
            Name of column according to values of which to color points in the plot.
            Can be both numeric and categorical. By default, all points have the same color.
        interactive : bool (default = True)
            Whether to show an interactive plot.
        standardize : bool (default = True)
            Whether to standardize the data to zero mean and unit variance before applying PCA.
        figsize : Tuple[float, float] (default = (12, 12))
            Size of the plot.
        opacity : float (default = 0.8)
            Opacity of the points in the plot. Higher means more opaque (less transparent).
        """
        df = df.dropna(subset=columns)

        pca = sklearn.decomposition.PCA(n_components=2)

        if standardize:
            scaler = StandardScaler()
            data_scaled = scaler.fit_transform(df[columns])
            pca_components = pca.fit_transform(data_scaled)
        else:
            pca_components = pca.fit_transform(df[columns])

        scatter_plot_2d(
            df=df,
            x=pca_components[:, 0],
            y=pca_components[:, 1],
            color_col=color_col,
            interactive=interactive,
            figsize=figsize,
            opacity=opacity,
            xlabel="First principal component",
            ylabel="Second principal component",
            show_xticks=True,
            show_yticks=True,
            show_zerolines=True,
            equal_scale_axes=True,
        )

        print(f"Explained variance ratio: {pca.explained_variance_ratio_[:2].sum() * 100 :.2f}%")

    @staticmethod
    def pca_explained_variance(
        df: pd.DataFrame,
        columns: Optional[List[str]] = None,
        standardize: bool = True,
        show_grid: bool = True,
        figsize: Tuple[float, float] = (10, 7),
    ) -> None:
        """Plot a plot of variance explained by each principal component.

        Parameters
        ----------
        df : pd.DataFrame
            Data on which to perform PCA.
        columns : List[str], optional
            Which columns to perform PCA on. All columns will be used by default.
        standardize : bool (default = True)
            Whether to standardize the data zero mean and unit variance before applying PCA.
        show_grid : bool (default = True)
            Whether to show a grid in the plot.
        figsize : Tuple[float, float] (default = (10, 7))
            Size of the plot.
        """
        df = df.dropna(subset=columns)

        pca = sklearn.decomposition.PCA()

        if standardize:
            scaler = StandardScaler()
            data_scaled = scaler.fit_transform(df[columns])
            pca.fit(data_scaled)
        else:
            pca.fit(df[columns])

        fig = plt.figure(figsize=figsize)
        plt.plot(pca.explained_variance_ratio_, figure=fig)
        plt.plot(np.cumsum(pca.explained_variance_ratio_), figure=fig)

        plt.legend(["Individual component", "Cumulative"])
        plt.xlabel("Principal component #")
        plt.ylabel("Explained variance ratio")
        plt.xticks(
            ticks=range(len(pca.explained_variance_ratio_)),
            labels=range(1, (len(pca.explained_variance_ratio_) + 1)),
        )
        if show_grid:
            plt.grid()
        plt.show()

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                total_dedent(
                    """
                    from edvart.report_sections.multivariate_analysis import PCA
                    pca_first_vs_second = PCA.pca_first_vs_second
                    pca_explained_variance = PCA.pca_explained_variance
                    """
                )
            ]
        return [
            "from edvart.plots import scatter_plot_2d",
            "import matplotlib.pyplot as plt",
            "import numpy as np",
            "import sklearn.decomposition",
            "from sklearn.preprocessing import StandardScaler",
        ]

    def add_cells(self, cells: List[Dict[str, Any]]) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        first_vs_second_header = nbfv4.new_markdown_cell("### First vs second principal component")
        cells.append(first_vs_second_header)
        explained_variance_header = nbfv4.new_markdown_cell("### Explained variance ratio")

        first_vs_second_call = "pca_first_vs_second(df=df"
        if not self.interactive:
            first_vs_second_call += ", interactive=False"
        explained_variance_call = "pca_explained_variance(df=df"
        if self.columns is not None:
            first_vs_second_call += f", columns={self.columns}"
            explained_variance_call += f", columns={self.columns}"
        if not self.standardize:
            explained_variance_call += ", standardize=False"
            first_vs_second_call += ", standardize=False"
        if not self.interactive:
            first_vs_second_call += ", interactive=False"
        if self.color_col is not None:
            first_vs_second_call += f", color_col='{self.color_col}'"
        first_vs_second_call += ")"
        explained_variance_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            cells.append(nbfv4.new_code_cell(first_vs_second_call))
            cells.append(explained_variance_header)
            cells.append(nbfv4.new_code_cell(explained_variance_call))
        else:
            first_vs_second_code = get_code(PCA.pca_first_vs_second) + "\n\n" + first_vs_second_call
            cells.append(nbfv4.new_code_cell(first_vs_second_code))

            cells.append(explained_variance_header)
            exaplained_variance_code = (
                get_code(PCA.pca_explained_variance) + "\n\n" + explained_variance_call
            )
            cells.append(nbfv4.new_code_cell(exaplained_variance_code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates the PCA section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        display(Markdown("### First vs second principal component"))
        PCA.pca_first_vs_second(
            df=df,
            columns=self.columns,
            color_col=self.color_col,
            interactive=self.interactive,
            standardize=self.standardize,
        )
        display(Markdown("### Explained variance ratio"))
        PCA.pca_explained_variance(df=df, columns=self.columns, standardize=self.standardize)


class ParallelCoordinates(Section):
    """Generates the Parallel coordinates subsection.

    Parameters
    ----------
    df : pd.DataFrame
        Data for which to generate the parallel coordinates plot.
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns for which to generate parallel coordinates. All columns which are either numeric or
        categorical with at most `nunique_max` unique values are used by default.
    nunique_max : int (default = 20)
        Maximum number of unqiue values for non-numeric columns to be plotted.
        Ignored if set to a negative number.
    color_col : str, optional
        Name of column determining color of the coordinate lines.
        Both numeric and categorical columns are supported.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        nunique_max: int = 20,
        color_col: Optional[str] = None,
    ):
        self.color_col = color_col
        if columns is None:
            columns = [
                col
                for col in df.columns
                if is_categorical(df[col], unique_value_count_threshold=nunique_max)
                or is_boolean(df[col])
                or is_numeric(df[col])
            ]
            # If all columns are numeric we don't want to list them all in the generated call
            # Setting columns to None will result in the columns argument not being included
            # instead of showing a potentially long list of all columns
            # in the generated call, therefore the default (all columns) will be used
            if len(columns) == len(df.columns):
                columns = None

        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Parallel coordinates"

    @staticmethod
    def parallel_coordinates(
        df: pd.DataFrame,
        columns: Optional[List[str]] = None,
        hide_columns: Optional[List[str]] = None,
        drop_na: bool = False,
        color_col: Optional[str] = None,
        show_colorscale: bool = True,
    ) -> None:
        """Generate the parallel coordinates interactive plot.

        Parameters
        ----------
        df : pd.DataFrame
            Data for which to generate the parallel coordinates plot.
        columns : List[str], optional
            Columns for which to generate the plot. All columns are used by default.
        hide_columns : List[str], optional
            Columns to exclude from plotting.
        drop_na : bool (default = False)
            Whether to drop NaNs in data.
        color_col : str, optional
            Which column to use for coloring of lines. Can be both numeric and categorical.
        show_colorscale : bool (default = True)
            Whether to show a color scale on the right side of the plot.
        """
        if columns is None:
            columns = df.columns
        if hide_columns is not None:
            columns = [col for col in columns if col not in hide_columns]
        if drop_na:
            df = df.dropna()
        if color_col is not None:
            is_categorical_color = not is_numeric(df[color_col])

            if is_categorical_color:
                categories = df[color_col].unique()
                colorscale = list(discrete_colorscale(len(categories)))
                # encode categories into numbers
                color_series = pd.Series(pd.Categorical(df[color_col]).codes)
            else:
                color_series = df[color_col]
                colorscale = "Bluered_r"

            line = {
                "color": color_series,
                "colorscale": colorscale,
                "showscale": show_colorscale,
                "colorbar": {"title": color_col, "lenmode": "pixels", "len": 300},
            }

            if is_categorical_color:
                line["colorbar"].update(
                    {
                        "tickvals": color_series.unique(),
                        "ticktext": categories,
                        "lenmode": "pixels",
                        "len": min(40 * len(categories), 300),
                    }
                )
        else:
            line = None

        numeric_columns = [col for col in columns if is_numeric(df[col])]
        categorical_columns = [col for col in columns if not is_numeric(df[col])]

        # Add numeric columns to dimensions
        dimensions = [{"label": col_name, "values": df[col_name]} for col_name in numeric_columns]
        # Add categorical columns to dimensions
        for col_name in categorical_columns:
            categories = df[col_name].unique()
            values = pd.Series(pd.Categorical(df[col_name]).codes)
            dimensions.append(
                {
                    "label": col_name,
                    "values": values,
                    "tickvals": values.unique(),
                    "ticktext": categories,
                }
            )

        fig = go.Figure(go.Parcoords(line=line, dimensions=dimensions))
        fig.show()

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                total_dedent(
                    """
                    from edvart.report_sections.multivariate_analysis import ParallelCoordinates
                    parallel_coordinates = ParallelCoordinates.parallel_coordinates
                    """
                )
            ]
        return [
            "from edvart.utils import discrete_colorscale",
            "import plotly",
            "import plotly.graph_objects as go",
            "plotly.offline.init_notebook_mode()",
        ]

    def add_cells(self, cells: List[Dict[str, Any]]) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "parallel_coordinates(df=df"
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        if self.color_col is not None:
            default_call += f", color_col='{self.color_col}'"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(discrete_colorscale)
                + "\n\n"
                + get_code(ParallelCoordinates.parallel_coordinates)
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates the Parallel coordinates section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        ParallelCoordinates.parallel_coordinates(
            df=df, columns=self.columns, color_col=self.color_col
        )


class ParallelCategories(Section):
    """
    Generates the Parallel categories subsection.

    Parameters
    ----------
    df : pd.DataFrame
        Data for which to generate the parallel coordinates plot.
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns for which to generate parallel coordinates.
        All categorical columns with at most `nunique_max` unique values are used by default.
    nunique_max : int (default = 20)
        Maximum number of unqiue values for non-numeric columns to be plotted.
        Ignored if set to a negative number.
    color_col : str, optional
        Name of column determining colors within categories.
        Both numeric and categorical columns are supported.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        nunique_max: int = 20,
        color_col: Optional[str] = None,
    ):
        self.color_col = color_col
        if columns is None:
            columns = [
                col
                for col in df.columns
                if is_categorical(df[col], unique_value_count_threshold=nunique_max)
                or is_boolean(df[col])
            ]

            # If all columns are numeric we don't want to list them all in the generated call
            # Setting columns to None will result in the columns argumented not being included
            # instead of showing a potentially long list of all columns
            # in the generated call, therefore the default (all columns) will be used
            if len(columns) == len(df.columns):
                columns = None

        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Parallel categories"

    @staticmethod
    def parallel_categories(
        df: pd.DataFrame,
        columns: Optional[List[str]] = None,
        hide_columns: Optional[List[str]] = None,
        drop_na: bool = False,
        color_col: Optional[str] = None,
    ) -> None:
        """Generate the parallel coordinates interactive plot.

        Parameters
        ----------
        df : pd.DataFrame
            Data for which to generate the parallel coordinates plot.
        columns : List[str], optional
            Columns for which to generate the plot.
        hide_columns : List[str], optional
            Columns to exclude from plotting.
        drop_na : bool (default = False)
            Whether to drop NaNs in data.
        color_col : str, optional
            Which column to use for coloring of lines. Can be both numeric and categorical.
        """
        if columns is None:
            columns = df.columns
        if hide_columns is not None:
            columns = [col for col in columns if col not in hide_columns]
        if drop_na:
            df = df.dropna()
        if color_col is not None:
            categorical_color = not is_numeric(df[color_col])
            if categorical_color:
                categories = df[color_col].unique()
                colorscale = list(discrete_colorscale(len(categories)))
                # encode categories into numbers
                color_series = pd.Series(pd.Categorical(df[color_col]).codes)
            else:
                color_series = df[color_col]
                colorscale = "Bluered_r"

            line = {
                "color": color_series,
                "colorscale": colorscale,
                "colorbar": {"title": color_col},
            }

            if categorical_color:
                line["colorbar"].update(
                    {
                        "tickvals": color_series.unique(),
                        "ticktext": categories,
                        "lenmode": "pixels",
                        "len": min(40 * len(categories), 300),
                    }
                )
        else:
            line = None

        dimensions = [
            go.parcats.Dimension(values=df[col_name], label=col_name) for col_name in columns
        ]

        fig = go.Figure(go.Parcats(dimensions=dimensions, line=line))
        fig.show()

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                total_dedent(
                    """
                    from edvart.report_sections.multivariate_analysis import ParallelCategories
                    parallel_categories = ParallelCategories.parallel_categories
                    """
                )
            ]
        return [
            "from edvart.utils import discrete_colorscale",
            "import plotly",
            "import plotly.graph_objects as go",
            "plotly.offline.init_notebook_mode()",
        ]

    def add_cells(self, cells: List[Dict[str, Any]]) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "parallel_categories(df=df"
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        if self.color_col is not None:
            default_call += f", color_col='{self.color_col}'"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(discrete_colorscale)
                + "\n\n"
                + get_code(ParallelCategories.parallel_categories)
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates the Parallel coordinates section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        ParallelCategories.parallel_categories(
            df=df, columns=self.columns, color_col=self.color_col
        )
